from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, Callable, Literal, NamedTuple, Protocol, overload, runtime_checkable

if TYPE_CHECKING:
    from cartographer.stream import Session

HomingAxis = Literal["x", "y", "z"]


@dataclass(frozen=True)
class Position:
    x: float
    y: float
    z: float

    def as_tuple(self) -> tuple[float, float, float]:
        return (self.x, self.y, self.z)

    def as_list(self) -> list[float]:
        return list(self.as_tuple())


class HomingState(Protocol):
    def is_homing_z(self) -> bool:
        """Check if the z axis is currently being homed."""
        ...

    def set_z_homed_position(self, position: float) -> None:
        """Set the homed position for the z axis."""
        ...


class Endstop(Protocol):
    """Endstop interface for homing operations."""

    def query_is_triggered(self, print_time: float) -> bool:
        """Return true if endstop is currently triggered"""
        ...

    def home_start(self, print_time: float) -> object:
        """Start the homing process"""
        ...

    def home_wait(self, home_end_time: float) -> float:
        """Wait for homing to complete"""
        ...

    def on_home_end(self, homing_state: HomingState) -> None:
        """To be called when the homing process is complete"""
        ...

    def get_endstop_position(self) -> float:
        """The position of the endstop on the rail"""
        ...


@dataclass(frozen=True)
class Sample:
    frequency: float
    time: float
    position: Position | None
    velocity: float | None
    temperature: float


class Mcu(Protocol):
    def start_homing_scan(self, print_time: float, frequency: float) -> object: ...
    def start_homing_touch(self, print_time: float, threshold: int) -> object: ...
    def stop_homing(self, home_end_time: float) -> float: ...
    def start_session(self, start_condition: Callable[[Sample], bool] | None = None) -> Session[Sample]: ...


class MacroParams(Protocol):
    @overload
    def get(self, name: str, default: str = ...) -> str: ...
    @overload
    def get(self, name: str, default: None) -> str | None: ...
    @overload
    def get_float(
        self, name: str, default: float = ..., *, above: float = ..., minval: float = ..., maxval: float = ...
    ) -> float: ...
    @overload
    def get_float(
        self, name: str, default: None, *, above: float = ..., minval: float = ..., maxval: float = ...
    ) -> float | None: ...
    def get_int(
        self,
        name: str,
        default: int = ...,
        *,
        minval: int = ...,
        maxval: int = ...,
    ) -> int: ...


@runtime_checkable
class SupportsFallbackMacro(Protocol):
    def set_fallback_macro(self, macro: Macro) -> None: ...


class Macro(Protocol):
    description: str

    def run(self, params: MacroParams) -> None: ...


class ProbeMode(Protocol):
    @property
    def offset(self) -> Position: ...
    @property
    def is_ready(self) -> bool: ...
    @property
    def last_homing_time(self) -> float: ...
    def get_status(self, eventtime: float) -> object: ...
    def perform_probe(self) -> float: ...


class TemperatureStatus(NamedTuple):
    current: float
    target: float


class Toolhead(Protocol):
    def get_last_move_time(self) -> float:
        """Returns the last time the toolhead moved."""
        ...

    def wait_moves(self) -> None:
        """Wait for all moves to complete."""
        ...

    def get_position(self) -> Position:
        """Get the currently commanded position of the toolhead."""
        ...

    def move(self, *, x: float | None = None, y: float | None = None, z: float | None = None, speed: float) -> None:
        """Move to requested position."""
        ...

    def is_homed(self, axis: HomingAxis) -> bool:
        """Check if axis is homed."""
        ...

    def get_gcode_z_offset(self) -> float:
        """Returns currently applied gcode offset for the z axis."""
        ...

    def z_homing_move(self, endstop: Endstop, *, speed: float) -> float:
        """Starts homing move towards the given endstop."""
        ...

    def set_z_position(self, z: float) -> None:
        """Set the z position of the toolhead."""
        ...

    def get_axis_limits(self, axis: HomingAxis) -> tuple[float, float]:
        """Get the limits of an axis."""
        ...

    def manual_probe(self, finalize_callback: Callable[[Position | None], None]) -> None:
        """Start a manual probe."""
        ...

    def clear_z_homing_state(self) -> None:
        """Clears z homing state"""
        ...

    def dwell(self, seconds: float) -> None:
        """Dwell for the given number of seconds."""
        ...

    def get_extruder_temperature(self) -> TemperatureStatus:
        """Get the current and target temperature of the extruder."""
        ...

    def apply_axis_twist_compensation(self, position: Position) -> Position:
        """Apply axis twist compensation to the given position."""
        ...



from __future__ import annotations

import logging
import time
from dataclasses import dataclass
from typing import TYPE_CHECKING, Literal, Protocol, final

import numpy as np
from typing_extensions import override

from cartographer.interfaces.printer import Macro, MacroParams, Toolhead

if TYPE_CHECKING:
    from cartographer.interfaces.configuration import Configuration
    from cartographer.probe import Probe

logger = logging.getLogger(__name__)


@dataclass
class CalibrationOptions:
    start: float | None
    end: float | None
    line: float | None


@dataclass
class CompensationResult:
    axis: Literal["x", "y"]
    start: float
    end: float
    values: list[float]


class AxisTwistCompensationAdapter(Protocol):
    move_height: float
    speed: float

    def clear_compensations(self, axis: Literal["x", "y"]) -> None: ...
    def apply_compensation(self, result: CompensationResult) -> None: ...
    def get_calibration_options(self, axis: Literal["x", "y"]) -> CalibrationOptions: ...


@final
class AxisTwistCompensationMacro(Macro):
    description = "Scan and touch to calculate axis twist compensation values."

    def __init__(
        self, probe: Probe, toolhead: Toolhead, adapter: AxisTwistCompensationAdapter, config: Configuration
    ) -> None:
        self.probe = probe
        self.toolhead = toolhead
        self.adapter = adapter
        self.config = config

    @override
    def run(self, params: MacroParams) -> None:
        axis = params.get("AXIS", default="x").lower()
        if axis not in ("x", "y"):
            msg = f"Invalid axis '{axis}'"
            raise RuntimeError(msg)
        sample_count = params.get_int("SAMPLE_COUNT", default=5)

        start_pos, end_pos, line_pos = self._get_calibration_positions(params, axis)

        self.adapter.clear_compensations(axis)
        try:
            self._calibrate(axis, sample_count, start_pos, end_pos, line_pos)
        except RuntimeError:
            logger.info("""
                Error during axis twist compensation calibration,
                existing compensation has been cleared.
                Restart firmware to restore.
                """)
            raise

    def _get_calibration_positions(self, params: MacroParams, axis: Literal["x", "y"]) -> tuple[float, float, float]:
        start_pos = params.get_float("START", default=None)
        end_pos = params.get_float("END", default=None)
        line_pos = params.get_float("LINE", default=None)

        options = self.adapter.get_calibration_options(axis)
        boundaries = self.probe.touch.boundaries
        if axis == "x":
            if start_pos is None:
                start_pos = options.start or boundaries.min_x
            if end_pos is None:
                end_pos = options.end or boundaries.max_x
            if line_pos is None:
                line_pos = options.line or round((boundaries.max_y + boundaries.min_y) / 2, 2)
            if not boundaries.is_within(x=start_pos, y=line_pos):
                msg = f"Start position {start_pos} is outside of touch boundaries"
                raise RuntimeError(msg)
            if not boundaries.is_within(x=end_pos, y=line_pos):
                msg = f"End position {end_pos} is outside of touch boundaries"
                raise RuntimeError(msg)

        elif axis == "y":
            if start_pos is None:
                start_pos = options.start or boundaries.min_y
            if end_pos is None:
                end_pos = options.end or boundaries.max_y
            if line_pos is None:
                line_pos = options.line or round((boundaries.max_x + boundaries.min_x) / 2, 2)
            if not boundaries.is_within(x=line_pos, y=start_pos):
                msg = f"Start position {start_pos} is outside of touch boundaries"
                raise RuntimeError(msg)
            if not boundaries.is_within(x=line_pos, y=end_pos):
                msg = f"End position {end_pos} is outside of touch boundaries"
                raise RuntimeError(msg)

        return start_pos, end_pos, line_pos

    def _calibrate(
        self,
        axis: Literal["x", "y"],
        sample_count: int,
        start_pos: float,
        end_pos: float,
        line_pos: float,
    ) -> None:
        step = (end_pos - start_pos) / (sample_count - 1)
        results: list[float] = []
        start_time = time.time()
        for i in range(sample_count):
            position = start_pos + i * step
            self._move_probe_to(axis, position, line_pos)
            scan = self.probe.perform_scan()
            self._move_nozzle_to(axis, position, line_pos)
            touch = self.probe.perform_touch()
            result = scan - touch
            logger.debug("Offset at %:.2f: %.6f", position, result)
            results.append(result)
        logger.debug("Axis twist measurements completed in %.2f seconds", time.time() - start_time)

        avg = float(np.mean(results))
        results = [avg - x for x in results]

        self.adapter.apply_compensation(CompensationResult(axis=axis, start=start_pos, end=end_pos, values=results))
        logger.info("""
            Axis twist compensation state has been saved
            for the current session.  The SAVE_CONFIG command will
            update the printer config file and restart the printer.
            """)
        logger.info(
            "Touch %s axis twist compensation calibration complete: mean z_offset: %.6f, offsets: (%s)",
            axis.upper(),
            avg,
            ", ".join(f"{s:.6f}" for s in results),
        )

    def _move_nozzle_to(self, axis: Literal["x", "y"], position: float, line_pos: float) -> None:
        self.toolhead.move(z=self.adapter.move_height, speed=self.adapter.speed)
        if axis == "x":
            self.toolhead.move(
                x=position,
                y=line_pos,
                speed=self.adapter.speed,
            )
        else:
            self.toolhead.move(
                x=line_pos,
                y=position,
                speed=self.adapter.speed,
            )

    def _move_probe_to(self, axis: Literal["x", "y"], position: float, calibration_axis: float) -> None:
        if axis == "x":
            self.toolhead.move(
                x=position - self.config.general.x_offset,
                y=calibration_axis - self.config.general.y_offset,
                speed=self.adapter.speed,
            )
        else:
            self.toolhead.move(
                x=calibration_axis - self.config.general.x_offset,
                y=position - self.config.general.y_offset,
                speed=self.adapter.speed,
            )



from __future__ import annotations

import json
import os
import shlex
import subprocess
from datetime import datetime, timezone
from pathlib import Path
from typing import Any, TypedDict

from hatchling.builders.config import BuilderConfig
from hatchling.builders.hooks.plugin.interface import BuildHookInterface
from typing_extensions import override


def _run_git_command(cmd: str) -> str:
    prog = shlex.split(cmd)
    process = subprocess.Popen(prog, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    ret, _ = process.communicate()
    retcode = process.wait()
    if retcode == 0:
        return ret.strip().decode()
    return ""


def get_commit_sha(root: str) -> str:
    cmd = f"git -C {root} rev-parse HEAD"
    return _run_git_command(cmd)


def retrieve_git_version(root: str) -> str:
    cmd = f"git -C {root} describe --always --tags --long --dirty"
    return _run_git_command(cmd)


class ReleaseInfo(TypedDict):
    project_name: str
    package_name: str
    urls: dict[str, str]
    package_version: str
    git_version: str
    commit_sha: str
    build_time: str
    system_dependencies: dict[str, list[str]]


class Project(TypedDict):
    name: str
    urls: dict[str, str]


class CustomBuildHook(BuildHookInterface[BuilderConfig]):
    @override
    def initialize(self, version: str, build_data: dict[str, Any]) -> None:  # pyright: ignore[reportExplicitAny]
        project: Project | None = self.metadata.config.get("project")  # pyright: ignore[reportUnknownMemberType]
        if project is None:
            msg = "Project not found in metadata"
            raise ValueError(msg)

        build_time = datetime.now(timezone.utc)
        urls = project.get("urls", {})

        ref_name = os.getenv("GITHUB_REF_NAME")
        repository = os.getenv("GITHUB_REPOSITORY")
        if ref_name is not None and repository is not None:
            urls["changelog"] = f"https://github.com/{repository}/releases/tag/{ref_name}"

        data = ReleaseInfo(
            project_name=project.get("name"),
            package_name=self.metadata.name,  # pyright: ignore[reportUnknownMemberType]
            urls=urls,
            package_version=self.metadata.version,  # pyright: ignore[reportUnknownMemberType]
            git_version=retrieve_git_version(self.root),
            commit_sha=get_commit_sha(self.root),
            build_time=datetime.isoformat(build_time, timespec="seconds"),
            system_dependencies={},
        )

        out_file = Path(self.root, "release_info")
        with open(out_file, "w") as f:
            json.dump(data, f)

        build_data["extra_metadata"][str(out_file)] = "release_info"

    @override
    def clean(self, versions: list[str]) -> None:
        out_file = Path(self.root, "release_info")
        out_file.unlink(missing_ok=True)



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Callable, final

from cartographer.adapters.klipper.mcu import KlipperCartographerMcu

if TYPE_CHECKING:
    from configfile import ConfigWrapper

    from cartographer.interfaces.printer import Sample

REPORT_TIME = 0.300

logger = logging.getLogger(__name__)


@final
class PrinterTemperatureCoil:
    def __init__(self, config: ConfigWrapper):
        self.printer = config.get_printer()
        self.name = config.get_name()
        self.min_temp = float("-inf")
        self.max_temp = float("inf")
        self.temperature_callback = None
        self.printer.register_event_handler("klippy:mcu_identify", self._handle_mcu_identify)

    def _handle_mcu_identify(self) -> None:
        carto = self.printer.lookup_object("cartographer")
        if not isinstance(carto.mcu, KlipperCartographerMcu):
            logger.error("Expected cartographer MCU to be of type KlipperCartographerMcu, got %s", type(carto.mcu))
            return
        carto.mcu.register_callback(self._sample_callback)

    def setup_callback(self, temperature_callback: Callable[[float, float], None]) -> None:
        self.temperature_callback = temperature_callback

    def get_report_time_delta(self) -> float:
        return REPORT_TIME

    def setup_minmax(self, min_temp: float, max_temp: float) -> None:
        self.min_temp = min_temp
        self.max_temp = max_temp

    def _sample_callback(self, sample: Sample) -> None:
        if self.temperature_callback is None:
            return
        self.temperature_callback(sample.time, sample.temperature)
        if not (self.min_temp <= sample.temperature <= self.max_temp):
            logger.warning(
                "temperature for %(sensor_name)s at %(temperature)s is out of range [%(min_temp)s, %(max_temp)s]",
                dict(
                    sensor_name=self.name,
                    temperature=sample.temperature,
                    min_temp=self.min_temp,
                    max_temp=self.max_temp,
                ),
            )
        return



from __future__ import annotations

import itertools
import logging
from typing import TYPE_CHECKING, cast

from pytest_bdd import given, parsers, scenarios, then, when

if TYPE_CHECKING:
    from unittest.mock import Mock

    from pytest import LogCaptureFixture
    from pytest_mock import MockerFixture

    from cartographer.interfaces.printer import MacroParams, Toolhead
    from cartographer.probe import Probe


scenarios("../../features/probe.feature")


@given("the probe measures:")
def given_probe_measurements(mocker: MockerFixture, probe: Probe, datatable: list[list[str]]):
    probe.scan.perform_probe = mocker.Mock(
        side_effect=itertools.cycle(map(float, itertools.chain.from_iterable(datatable)))
    )


@given("the probe is triggered")
def given_probe_triggered(mocker: MockerFixture, probe: Probe):
    probe.scan.query_is_triggered = mocker.Mock(return_value=True)


@given("the probe is not triggered")
def given_probe_not_triggered(mocker: MockerFixture, probe: Probe):
    probe.scan.query_is_triggered = mocker.Mock(return_value=False)


@when("I run the PROBE macro")
def when_run_probe_macro(params: MacroParams, caplog: LogCaptureFixture, probe: Probe):
    from cartographer.macros.probe import ProbeMacro

    macro = ProbeMacro(probe)
    with caplog.at_level(logging.INFO):
        macro.run(params)


@when("I run the PROBE_ACCURACY macro")
def when_run_probe_accuracy_macro(params: MacroParams, caplog: LogCaptureFixture, probe: Probe, toolhead: Toolhead):
    from cartographer.macros.probe import ProbeAccuracyMacro

    macro = ProbeAccuracyMacro(probe, toolhead)
    with caplog.at_level(logging.INFO):
        macro.run(params)


@when("I run the QUERY_PROBE macro")
def when_run_query_probe_macro(params: MacroParams, caplog: LogCaptureFixture, probe: Probe):
    from cartographer.macros.probe import QueryProbeMacro

    macro = QueryProbeMacro(probe)
    with caplog.at_level(logging.INFO):
        macro.run(params)


@then(parsers.parse("it should probe {count:d} times"))
def then_probe_count(probe: Probe, count: int):
    assert cast("Mock", probe.scan.perform_probe).call_count == count



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Callable, TypedDict, final

import mcu
from mcu import MCU_trsync
from mcu import TriggerDispatch as KlipperTriggerDispatch
from typing_extensions import override

from cartographer.adapters.klipper.mcu.commands import (
    HomeCommand,
    KlipperCartographerCommands,
    ThresholdCommand,
    TriggerMethod,
)
from cartographer.adapters.klipper.mcu.constants import (
    FREQUENCY_RANGE_PERCENT,
    SHORTED_FREQUENCY_VALUE,
    TRIGGER_HYSTERESIS,
    KlipperCartographerConstants,
)
from cartographer.adapters.klipper.mcu.stream import KlipperStream, KlipperStreamMcu
from cartographer.interfaces.printer import Mcu, Position, Sample

if TYPE_CHECKING:
    from configfile import ConfigWrapper
    from reactor import ReactorCompletion

    from cartographer.stream import Session

logger = logging.getLogger(__name__)


class _RawData(TypedDict):
    clock: int
    data: int
    temp: int


@final
class KlipperCartographerMcu(Mcu, KlipperStreamMcu):
    _constants: KlipperCartographerConstants | None = None
    _commands: KlipperCartographerCommands | None = None

    @property
    def constants(self) -> KlipperCartographerConstants:
        if self._constants is None:
            msg = "Mcu not initialized"
            raise RuntimeError(msg)
        return self._constants

    @property
    def commands(self) -> KlipperCartographerCommands:
        if self._commands is None:
            msg = "Mcu not initialized"
            raise RuntimeError(msg)
        return self._commands

    def __init__(
        self,
        config: ConfigWrapper,
        smoothing_fn: Callable[[Sample], Sample] | None = None,
    ):
        self.printer = config.get_printer()
        self.klipper_mcu = mcu.get_printer_mcu(self.printer, config.get("mcu"))
        self._stream = KlipperStream[Sample](self, self.klipper_mcu.get_printer().get_reactor(), smoothing_fn)
        self.dispatch = KlipperTriggerDispatch(self.klipper_mcu)

        self.motion_report = self.printer.load_object(config, "motion_report")

        self.printer.register_event_handler("klippy:mcu_identify", self._handle_mcu_identify)
        self.printer.register_event_handler("klippy:connect", self._handle_connect)
        self.printer.register_event_handler("klippy:shutdown", self._handle_shutdown)
        self.klipper_mcu.register_config_callback(self._initialize)

    def _initialize(self) -> None:
        self._constants = KlipperCartographerConstants(self.klipper_mcu)
        self._commands = KlipperCartographerCommands(self.klipper_mcu)
        self.klipper_mcu.register_response(self._handle_data, "cartographer_data")

    @override
    def start_homing_scan(self, print_time: float, frequency: float) -> ReactorCompletion:
        self._set_threshold(frequency)
        completion = self.dispatch.start(print_time)

        self.commands.send_home(
            HomeCommand(
                trsync_oid=self.dispatch.get_oid(),
                trigger_reason=MCU_trsync.REASON_ENDSTOP_HIT,
                trigger_invert=0,
                threshold=0,
                trigger_method=TriggerMethod.SCAN,
            )
        )
        return completion

    @override
    def start_homing_touch(self, print_time: float, threshold: int) -> ReactorCompletion:
        completion = self.dispatch.start(print_time)

        self.commands.send_home(
            HomeCommand(
                trsync_oid=self.dispatch.get_oid(),
                trigger_reason=MCU_trsync.REASON_ENDSTOP_HIT,
                trigger_invert=0,
                threshold=threshold,
                trigger_method=TriggerMethod.TOUCH,
            )
        )
        return completion

    @override
    def stop_homing(self, home_end_time: float) -> float:
        self.dispatch.wait_end(home_end_time)
        self.commands.send_stop_home()
        result = self.dispatch.stop()
        if result >= MCU_trsync.REASON_COMMS_TIMEOUT:
            msg = "Communication timeout during homing"
            raise RuntimeError(msg)
        if result != MCU_trsync.REASON_ENDSTOP_HIT:
            return 0.0

        # TODO: Use a query state command for actual end time
        return home_end_time

    @override
    def start_session(self, start_condition: Callable[[Sample], bool] | None = None) -> Session[Sample]:
        return self._stream.start_session(start_condition)

    def register_callback(self, callback: Callable[[Sample], None]) -> None:
        return self._stream.register_callback(callback)

    @override
    def start_streaming(self) -> None:
        self.commands.send_stream_state(enable=True)

    @override
    def stop_streaming(self) -> None:
        self.commands.send_stream_state(enable=False)

    def _set_threshold(self, trigger_frequency: float) -> None:
        trigger = self.constants.frequency_to_count(trigger_frequency)
        untrigger = self.constants.frequency_to_count(trigger_frequency * (1 - TRIGGER_HYSTERESIS))

        self.commands.send_threshold(ThresholdCommand(trigger, untrigger))

    def _handle_mcu_identify(self) -> None:
        kin = self.printer.lookup_object("toolhead").get_kinematics()
        for stepper in kin.get_steppers():
            if stepper.is_active_axis("z"):
                self.dispatch.add_stepper(stepper)

    def _handle_connect(self) -> None:
        self.stop_streaming()

    def _handle_shutdown(self) -> None:
        self.stop_streaming()

    def _handle_data(self, data: _RawData) -> None:
        self._validate_data(data)
        clock = self.klipper_mcu.clock32_to_clock64(data["clock"])
        time = self.klipper_mcu.clock_to_print_time(clock)

        frequency = self.constants.count_to_frequency(data["data"])
        temperature = self.constants.calculate_temperature(data["temp"])
        position, velocity = self.get_requested_position(time)

        sample = Sample(time=time, frequency=frequency, temperature=temperature, position=position, velocity=velocity)
        self._stream.add_item(sample)

    _data_error: str | None = None

    def _validate_data(self, data: _RawData) -> None:
        count = data["data"]
        error: str | None = None
        if count == SHORTED_FREQUENCY_VALUE:
            error = "coil is shorted or not connected."
        elif count > self.constants.minimum_count * FREQUENCY_RANGE_PERCENT:
            error = "coil frequency reading exceeded max expected value, received %(data)d"

        if self._data_error == error:
            return
        self._data_error = error

        if error is None:
            return

        logger.error(error, {"data": data})
        if len(self._stream.sessions) > 0:
            self.klipper_mcu.get_printer().invoke_shutdown(error % {"data": data})

    def get_requested_position(self, time: float) -> tuple[Position | None, float | None]:
        trapq = self.motion_report.trapqs.get("toolhead")
        if trapq is None:
            logger.warning("No dump trapq for toolhead, cannot get position at time %.3f", time)
            return None, None
        position, velocity = trapq.get_trapq_position(time)
        if position is None:
            return None, velocity
        return Position(x=position[0], y=position[1], z=position[2]), velocity



import logging
import time
from functools import wraps
from typing import Callable, TypeVar

from typing_extensions import ParamSpec

P = ParamSpec("P")
R = TypeVar("R")


def log_duration(message: str) -> Callable[[Callable[P, R]], Callable[P, R]]:
    def decorator(func: Callable[P, R]) -> Callable[P, R]:
        @wraps(func)
        def wrapper(*args: P.args, **kwargs: P.kwargs) -> R:
            start_time = time.time()
            try:
                return func(*args, **kwargs)
            finally:
                duration = time.time() - start_time
                func_logger = logging.getLogger(func.__module__)
                func_logger.debug("%s completed in %.2f seconds", message, duration)

        return wrapper

    return decorator



from __future__ import annotations

import logging
from typing import final

logger = logging.getLogger(__name__)
MIN_DT = 1e-4


@final
class AlphaBetaFilter:
    def __init__(self, alpha: float = 0.5, beta: float = 1e-6):
        if not (0 <= alpha <= 1 and 0 <= beta <= 1):
            msg = "Alpha and beta must be between 0 and 1"
            raise ValueError(msg)

        self.alpha = alpha
        self.beta = beta

        self.position: float | None = None  # last estimated position
        self.velocity: float = 0.0  # last estimated velocity
        self.last_time: float = 0.0  # time of last update

    def update(self, *, measurement: float, time: float) -> float:
        if self.position is None:
            self.position = measurement
            self.last_time = time
            return self.position

        dt = time - self.last_time
        self.last_time = time  # Update time even if out-of-order, to avoid stalling

        # Predict the next position
        predicted_position = self.position + self.velocity * dt

        # Calculate residual (difference from predicted to actual)
        residual = measurement - predicted_position

        # Always update position
        self.position = predicted_position + self.alpha * residual

        # Update velocity only if dt > 0
        if dt > MIN_DT:
            self.velocity = self.velocity + (self.beta * residual) / dt
        else:
            logger.debug("Skipping velocity update due to tiny or negative dt: %.6f", dt)

        return self.position



from __future__ import annotations

from typing import TYPE_CHECKING, final

from gcode import CommandError

from cartographer.adapters.utils import reraise_as

if TYPE_CHECKING:
    from gcode import GCodeCommand

    from cartographer.adapters.klipper.toolhead import KlipperToolhead
    from cartographer.interfaces.printer import ProbeMode
    from cartographer.macros.probe import ProbeMacro, QueryProbeMacro


class KlipperProbeSession:
    def __init__(self, probe: ProbeMode, toolhead: KlipperToolhead) -> None:
        self._probe: ProbeMode = probe
        self._results: list[list[float]] = []
        self.toolhead: KlipperToolhead = toolhead

    @reraise_as(CommandError)
    def run_probe(self, gcmd: GCodeCommand) -> None:
        del gcmd
        pos = self.toolhead.get_position()
        trigger_pos = self._probe.perform_probe()
        self._results.append([pos.x, pos.y, trigger_pos])

    def pull_probed_results(self):
        result = self._results
        self._results = []
        return result

    def end_probe_session(self) -> None:
        self._results = []


@final
class KlipperCartographerProbe:
    def __init__(
        self,
        toolhead: KlipperToolhead,
        probe: ProbeMode,
        probe_macro: ProbeMacro,
        query_probe_macro: QueryProbeMacro,
    ) -> None:
        self.probe = probe
        self.probe_macro = probe_macro
        self.query_probe_macro = query_probe_macro
        self.toolhead = toolhead

    def get_probe_params(self, gcmd: GCodeCommand | None = None):
        del gcmd
        return {
            "probe_speed": 5,
            "lift_speed": 5,
            "samples": 1,
            "sample_retract_dist": 0.2,
            "samples_tolerance": 0.1,
            "samples_tolerance_retries": 0,
            "samples_result": "median",
        }

    def get_offsets(self) -> tuple[float, float, float]:
        return self.probe.offset.as_tuple()

    def get_status(self, eventtime: float):
        del eventtime
        return {
            "name": "cartographer",
            "last_query": 1 if self.query_probe_macro.last_triggered else 0,
            "last_z_result": self.probe_macro.last_trigger_position or 0,
        }

    def start_probe_session(self, gcmd: GCodeCommand) -> KlipperProbeSession:
        del gcmd
        return KlipperProbeSession(self.probe, self.toolhead)



from __future__ import annotations

import logging
from math import floor, sqrt
from typing import Generic, Protocol, TypeVar, final

logger = logging.getLogger(__name__)

MAX_CLUSTER_DISTANCE = 1.0
CELL_SIZE = 0.5  # must be ≤ MAX_CLUSTER_DISTANCE for full coverage


class Point(Protocol):
    x: float
    y: float


P = TypeVar("P", bound=Point, covariant=True)


@final
class NearestNeighborSearcher(Generic[P]):
    def __init__(self, points: list[P]) -> None:
        self.points = points
        self.grid: dict[tuple[int, int], list[P]] = {}
        for point in points:
            cell = self._cell_key(point.x, point.y)
            self.grid.setdefault(cell, []).append(point)

    def batch_query(self, positions: list[tuple[float, float]]) -> list[P | None]:
        results: list[P | None] = []

        for px, py in positions:
            best = None
            best_dist = MAX_CLUSTER_DISTANCE

            cx, cy = self._cell_key(px, py)
            for dx in (-1, 0, 1):
                for dy in (-1, 0, 1):
                    cell = (cx + dx, cy + dy)
                    for p in self.grid.get(cell, []):
                        dist = sqrt((px - p.x) ** 2 + (py - p.y) ** 2)
                        if dist < best_dist:
                            best_dist = dist
                            best = p
            results.append(best)

        return results

    def _cell_key(self, x: float, y: float) -> tuple[int, int]:
        return (floor(x / CELL_SIZE), floor(y / CELL_SIZE))



# https://github.com/Klipper3d/klipper/blob/master/klippy/extras/homing.py

from typing import Protocol

from mcu import MCU_endstop
from stepper import MCU_stepper

type _Pos = list[float]

class _McuProbe(Protocol):
    def get_steppers(self) -> list[MCU_stepper]: ...

class PrinterHoming:
    def probing_move(self, mcu_probe: _McuProbe, pos: _Pos, speed: float) -> _Pos: ...

class Homing:
    def set_homed_position(self, pos: list[float | None]) -> None: ...
    def get_axes(self) -> list[int]: ...

class HomingMove:
    def __init__(self, endstops: list[tuple[MCU_endstop, str]]) -> None: ...
    def homing_move(self, pos: _Pos, speed: float, probe_pos: bool) -> _Pos: ...



from __future__ import annotations

from dataclasses import dataclass
from typing import Literal, Protocol


@dataclass(frozen=True)
class GeneralConfig:
    x_offset: float
    y_offset: float
    z_backlash: float
    travel_speed: float
    verbose: bool
    macro_prefix: str | None


@dataclass(frozen=True)
class ScanConfig:
    samples: int
    models: dict[str, ScanModelConfiguration]
    probe_speed: float
    mesh_runs: int
    mesh_height: float
    mesh_direction: Literal["x", "y"]
    mesh_path: Literal["snake", "alternating_snake", "spiral", "random"]


@dataclass(frozen=True)
class TouchConfig:
    samples: int
    max_samples: int
    models: dict[str, TouchModelConfiguration]


@dataclass(frozen=True)
class BedMeshConfig:
    mesh_min: tuple[float, float]
    mesh_max: tuple[float, float]
    probe_count: tuple[int, int]
    speed: float
    horizontal_move_z: float
    adaptive_margin: float
    zero_reference_position: tuple[float, float]


@dataclass(frozen=True)
class ScanModelConfiguration:
    name: str
    coefficients: list[float]
    domain: tuple[float, float]
    z_offset: float


@dataclass(frozen=True)
class TouchModelConfiguration:
    name: str
    threshold: int
    speed: float
    z_offset: float


class Configuration(Protocol):
    general: GeneralConfig
    scan: ScanConfig
    touch: TouchConfig
    bed_mesh: BedMeshConfig

    def save_scan_model(self, config: ScanModelConfiguration) -> None: ...
    def save_touch_model(self, config: TouchModelConfiguration) -> None: ...
    def save_z_backlash(self, backlash: float) -> None: ...



from __future__ import annotations

import pytest

from cartographer.probe.touch_mode import TouchBoundaries, TouchModeConfiguration

BOUNDS = TouchBoundaries(min_x=10.0, max_x=20.0, min_y=5.0, max_y=15.0)


@pytest.mark.parametrize(
    "x, y, expected",
    [
        # Clearly within bounds
        (15.0, 10.0, True),
        (10.0, 5.0, True),  # on lower edge
        (20.0, 15.0, True),  # on upper edge
        # Clearly outside bounds (beyond tolerance)
        (9.98, 10.0, False),
        (20.02, 10.0, False),
        (15.0, 4.98, False),
        (15.0, 15.02, False),
        # Near boundary (within 0.01 tolerance)
        (9.99, 10.0, True),
        (20.01, 10.0, True),
        (15.0, 4.99, True),
        (15.0, 15.01, True),
    ],
)
def test_is_within_bounds(x: float, y: float, expected: bool) -> None:
    assert BOUNDS.is_within(x=x, y=y) is expected


def make_config(
    mesh_min: tuple[float, float],
    mesh_max: tuple[float, float],
    x_offset: float,
    y_offset: float,
) -> TouchModeConfiguration:
    return TouchModeConfiguration(
        samples=1,
        max_samples=1,
        mesh_min=mesh_min,
        mesh_max=mesh_max,
        x_offset=x_offset,
        y_offset=y_offset,
        models={},
    )


@pytest.mark.parametrize(
    "mesh_min, mesh_max, x_offset, y_offset, expected",
    [
        # Zero offsets
        ((0.0, 0.0), (100.0, 100.0), 0.0, 0.0, TouchBoundaries(min_x=0.0, max_x=100.0, min_y=0.0, max_y=100.0)),
        # Positive offsets
        ((0.0, 0.0), (100.0, 100.0), 10.0, 5.0, TouchBoundaries(min_x=0.0, max_x=90.0, min_y=0.0, max_y=95.0)),
        # Negative offsets
        ((0.0, 0.0), (100.0, 100.0), -10.0, -5.0, TouchBoundaries(min_x=10.0, max_x=100.0, min_y=5.0, max_y=100.0)),
        # Non-zero mesh_min
        ((10.0, 20.0), (100.0, 100.0), 5.0, 10.0, TouchBoundaries(min_x=10.0, max_x=95.0, min_y=20.0, max_y=90.0)),
    ],
)
def test_from_config_bounds(
    mesh_min: tuple[float, float],
    mesh_max: tuple[float, float],
    x_offset: float,
    y_offset: float,
    expected: TouchBoundaries,
) -> None:
    config = make_config(mesh_min, mesh_max, x_offset, y_offset)
    bounds = TouchBoundaries.from_config(config)
    assert bounds == expected



from __future__ import annotations

from dataclasses import replace
from typing import final

from typing_extensions import override

from cartographer.interfaces.configuration import (
    BedMeshConfig,
    Configuration,
    GeneralConfig,
    ScanConfig,
    ScanModelConfiguration,
    TouchConfig,
    TouchModelConfiguration,
)

default_general_config = GeneralConfig(
    x_offset=0.0,
    y_offset=0.0,
    travel_speed=300.0,
    z_backlash=0,
    macro_prefix="cartographer",
    verbose=False,
)
default_scan_config = ScanConfig(
    samples=20,
    models={},
    probe_speed=5.0,
    mesh_runs=1,
    mesh_direction="x",
    mesh_height=4.0,
    mesh_path="snake",
)
default_touch_config = TouchConfig(
    samples=5,
    max_samples=10,
    models={},
)
default_bed_mesh_config = BedMeshConfig(
    mesh_min=(0.0, 0.0),
    mesh_max=(200.0, 200.0),
    probe_count=(10, 10),
    speed=100,
    horizontal_move_z=3,
    adaptive_margin=2,
    zero_reference_position=(100, 100),
)


@final
class MockConfiguration(Configuration):
    def __init__(
        self,
        *,
        general: GeneralConfig | None = None,
        scan: ScanConfig | None = None,
        touch: TouchConfig | None = None,
        bed_mesh: BedMeshConfig | None = None,
    ):
        self.general = general or default_general_config
        self.scan = scan or default_scan_config
        self.touch = touch or default_touch_config
        self.bed_mesh = bed_mesh or default_bed_mesh_config

    @override
    def save_scan_model(self, config: ScanModelConfiguration) -> None:
        self.scan.models[config.name] = config

    @override
    def save_touch_model(self, config: TouchModelConfiguration) -> None:
        self.touch.models[config.name] = config

    @override
    def save_z_backlash(self, backlash: float) -> None:
        self.general = replace(self.general, z_backlash=backlash)



from __future__ import annotations

from typing import TYPE_CHECKING, Iterator, Literal, final

import numpy as np
from typing_extensions import override

from cartographer.macros.bed_mesh.interfaces import PathGenerator
from cartographer.macros.bed_mesh.paths.utils import (
    Vec,
    angle_deg,
    arc_points,
    cluster_points,
    normalize,
    perpendicular,
    row_direction,
)

if TYPE_CHECKING:
    from cartographer.macros.bed_mesh.interfaces import Point

BUFFER = 0.5


@final
class SnakePathGenerator(PathGenerator):
    def __init__(self, main_direction: Literal["x", "y"]):
        self.main_direction: Literal["x", "y"] = main_direction

    @override
    def generate_path(
        self,
        points: list[Point],
        x_axis_limits: tuple[float, float],
        y_axis_limits: tuple[float, float],
    ) -> Iterator[Point]:
        grid = cluster_points(points, self.main_direction)

        axis_min, axis_max = x_axis_limits if self.main_direction == "x" else y_axis_limits
        main_index = 0 if self.main_direction == "x" else 1
        secondary_index = 1 if self.main_direction == "x" else 0

        row_spacing = abs(grid[1][0][secondary_index] - grid[0][0][secondary_index])
        max_radius_by_spacing = round(row_spacing / 2, 2)

        mesh_min = grid[0][0][main_index]
        mesh_max = grid[-1][-1][main_index]
        max_radius_by_bounds = min(mesh_min - axis_min, axis_max - mesh_max) - BUFFER

        # Final corner radius
        corner_radius = float(max(0, min(max_radius_by_spacing, max_radius_by_bounds)))

        prev_row = grid[0]

        for i, row in enumerate(grid):
            row = list(row)
            if i % 2 == 1:
                row.reverse()

            if i > 0:
                prev_last = prev_row[-1]
                curr_first = row[0]
                entry_dir = row_direction(prev_row[-2:])  # You must define this separately

                yield from u_turn(prev_last, curr_first, entry_dir, corner_radius)

            yield from row
            prev_row = row


def u_turn(start: Point, end: Point, entry_dir: Vec, radius: float) -> Iterator[Point]:
    """Create two 90° arcs at each point for a smooth U-turn."""
    p1: Vec = np.array(start, dtype=float)
    p2: Vec = np.array(end, dtype=float)
    delta = p2 - p1

    if np.linalg.norm(delta) == 0:
        return  # skip zero-distance turn

    turn_dir = normalize(delta)

    # Determine if the turn is CCW or CW based on 2D cross product
    # We'll assume a horizontal travel: if moving left to right, then back right to left
    # Then the perpendicular should flip for the reverse direction
    turn_ccw = bool(np.cross(entry_dir, turn_dir) > 0)  # flip depending on entry direction
    turn_angle = 90 if turn_ccw else -90

    entry_perp = perpendicular(entry_dir, ccw=turn_ccw)
    start_angle = angle_deg(-entry_perp)

    offset = entry_perp * radius
    yield from arc_points(p1 + offset, radius, start_angle, turn_angle)
    yield from arc_points(p2 - offset, radius, start_angle + turn_angle, turn_angle)



from __future__ import annotations

import logging
from collections import defaultdict
from dataclasses import dataclass
from math import ceil
from typing import TYPE_CHECKING, Callable, Sequence, final

import numpy as np

from cartographer.interfaces.printer import Position, Sample

if TYPE_CHECKING:
    from numpy.typing import NDArray

    from cartographer.macros.bed_mesh.interfaces import Point

MIN_GRID_RESOLUTION = 3
DEFAULT_MAX_SAMPLE_DISTANCE = 1.0

logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class MeshGrid:
    """Represents a 2D mesh grid with coordinate mappings and bounds."""

    min_point: Point
    max_point: Point
    x_resolution: int
    y_resolution: int

    def __post_init__(self):
        if self.x_resolution < MIN_GRID_RESOLUTION or self.y_resolution < MIN_GRID_RESOLUTION:
            msg = f"Grid resolution must be at least {MIN_GRID_RESOLUTION}x{MIN_GRID_RESOLUTION}"
            raise ValueError(msg)

    @property
    def x_coords(self) -> NDArray[np.float64]:
        """Get array of x coordinates."""
        return np.round(np.linspace(self.min_point[0], self.max_point[0], self.x_resolution), 2)

    @property
    def y_coords(self) -> NDArray[np.float64]:
        """Get array of y coordinates."""
        return np.round(np.linspace(self.min_point[1], self.max_point[1], self.y_resolution), 2)

    @property
    def x_step(self) -> float:
        """Get step size in x direction."""
        return float(self.max_point[0] - self.min_point[0]) / (self.x_resolution - 1)

    @property
    def y_step(self) -> float:
        """Get step size in y direction."""
        return float(self.max_point[1] - self.min_point[1]) / (self.y_resolution - 1)

    def generate_points(self) -> list[Point]:
        """Generate all grid points in y-major order."""
        return [(float(x), float(y)) for x in self.x_coords for y in self.y_coords]

    def contains_point(self, point: Point, epsilon: float = 1e-2) -> bool:
        """Check if a point is within the grid bounds."""
        x, y = point
        return bool(
            self.min_point[0] - epsilon <= x <= self.max_point[0] + epsilon
            and self.min_point[1] - epsilon <= y <= self.max_point[1] + epsilon
        )

    def point_to_grid_index(self, point: Point) -> tuple[int, int]:
        """Convert a point to grid indices (j, i) where j=row, i=col."""
        x, y = point
        i = round((x - self.min_point[0]) / self.x_step)
        j = round((y - self.min_point[1]) / self.y_step)
        return j, i

    def grid_index_to_point(self, j: int, i: int) -> Point:
        """Convert grid indices to a point."""
        x = self.x_coords[i]
        y = self.y_coords[j]
        return (float(x), float(y))

    def is_valid_index(self, j: int, i: int) -> bool:
        """Check if grid indices are valid."""
        return 0 <= i < self.x_resolution and 0 <= j < self.y_resolution


@dataclass(frozen=True)
class GridPointResult:
    """Result of assigning samples to a grid point."""

    point: Point
    z: float
    sample_count: int


@final
class SampleProcessor:
    """Handles processing and assignment of samples to grid points."""

    def __init__(self, grid: MeshGrid, max_distance: float = DEFAULT_MAX_SAMPLE_DISTANCE):
        self.grid = grid
        self.max_distance = max_distance

    def assign_samples_to_grid(
        self,
        samples: list[Sample],
        calculate_height: Callable[[Sample], float],
    ) -> list[GridPointResult]:
        """Assign samples to grid points and calculate median heights."""
        accumulator: dict[tuple[int, int], list[float]] = defaultdict(list)
        sample_points = [
            (sample.position.x, sample.position.y, sample) for sample in samples if sample.position is not None
        ]

        for x, y, sample in sample_points:
            if not self.grid.contains_point((x, y)):
                continue

            j, i = self.grid.point_to_grid_index((x, y))
            if not self.grid.is_valid_index(j, i):
                continue

            grid_point = self.grid.grid_index_to_point(j, i)
            distance = np.hypot(x - grid_point[0], y - grid_point[1])
            if distance > self.max_distance:
                continue

            sample_height = calculate_height(sample)
            accumulator[(j, i)].append(sample_height)

        results: list[GridPointResult] = []
        for j in range(self.grid.y_resolution):
            for i in range(self.grid.x_resolution):
                grid_point = self.grid.grid_index_to_point(j, i)
                values = accumulator.get((j, i), [])
                count = len(values)

                z = float(np.median(values)) if values else np.nan
                results.append(GridPointResult(point=grid_point, z=z, sample_count=count))

        return results


@final
class CoordinateTransformer:
    """Handles coordinate transformations between different reference frames."""

    def __init__(self, probe_offset: Position):
        self.probe_offset = probe_offset

    def probe_to_nozzle(self, point: Point) -> Point:
        """Convert probe coordinates to nozzle coordinates."""
        x, y = point
        return (x - self.probe_offset.x, y - self.probe_offset.y)

    def nozzle_to_probe(self, point: Point) -> Point:
        """Convert nozzle coordinates to probe coordinates."""
        x, y = point
        return (x + self.probe_offset.x, y + self.probe_offset.y)

    def normalize_to_zero_reference(self, positions: list[Position], zero_ref: Point) -> list[Position]:
        """Normalize positions to a zero reference point using bilinear interpolation."""
        if not positions:
            return []

        # Extract sorted unique coordinates
        xs = np.array(sorted({p.x for p in positions}))
        ys = np.array(sorted({p.y for p in positions}))

        # Build a mapping from (x, y) to z
        z_map = {(p.x, p.y): p.z for p in positions}

        # Build Z grid with shape (y_count, x_count)
        z_grid = np.array([[z_map[(x, y)] for x in xs] for y in ys])

        # Compute reference height using bilinear interpolation
        z_ref = self._bilinear_interpolate(xs, ys, z_grid, zero_ref)

        # Normalize all heights
        z_grid -= z_ref

        # Rebuild list in y-major order
        normalized_positions = [
            Position(x=float(x), y=float(y), z=float(z_grid[yi, xi]))
            for yi, y in enumerate(ys)
            for xi, x in enumerate(xs)
        ]
        return normalized_positions

    def _bilinear_interpolate(
        self,
        xs: NDArray[np.float64],
        ys: NDArray[np.float64],
        z_grid: NDArray[np.float64],
        point: Point,
    ) -> float:
        """Perform bilinear interpolation to find height at a point."""
        dx = xs[1] - xs[0]
        dy = ys[1] - ys[0]

        tx = (point[0] - xs[0]) / dx
        ty = (point[1] - ys[0]) / dy
        xi = int(np.floor(tx))
        yi = int(np.floor(ty))
        tx -= xi
        ty -= yi

        z00 = z_grid[yi, xi]
        z01 = z_grid[yi, xi + 1]
        z10 = z_grid[yi + 1, xi]
        z11 = z_grid[yi + 1, xi + 1]

        z0 = (1 - tx) * z00 + tx * z01
        z1 = (1 - tx) * z10 + tx * z11
        z_ref = (1 - ty) * z0 + ty * z1

        return float(z_ref)


@dataclass(frozen=True)
class MeshBounds:
    """Represents the bounds of a mesh."""

    min_point: Point
    max_point: Point

    def width(self) -> float:
        return float(self.max_point[0] - self.min_point[0])

    def height(self) -> float:
        return float(self.max_point[1] - self.min_point[1])


@final
class AdaptiveMeshCalculator:
    """Calculates adaptive mesh bounds and resolution."""

    def __init__(self, base_bounds: MeshBounds, base_resolution: tuple[int, int]):
        self.base_bounds = base_bounds
        self.base_resolution = base_resolution

    def calculate_adaptive_bounds(self, object_points: Sequence[Point], margin: float) -> MeshBounds:
        """Calculate adaptive bounds based on object points."""
        if not object_points:
            return self.base_bounds

        min_x = min(x for x, _ in object_points)
        max_x = max(x for x, _ in object_points)
        min_y = min(y for _, y in object_points)
        max_y = max(y for _, y in object_points)

        adapted_min_x = max(min_x - margin, self.base_bounds.min_point[0])
        adapted_max_x = min(max_x + margin, self.base_bounds.max_point[0])
        adapted_min_y = max(min_y - margin, self.base_bounds.min_point[1])
        adapted_max_y = min(max_y + margin, self.base_bounds.max_point[1])

        return MeshBounds(min_point=(adapted_min_x, adapted_min_y), max_point=(adapted_max_x, adapted_max_y))

    def calculate_adaptive_resolution(self, adaptive_bounds: MeshBounds) -> tuple[int, int]:
        """Calculate resolution maintaining point density."""
        base_width = self.base_bounds.width()
        base_height = self.base_bounds.height()

        if base_width == 0 or base_height == 0:
            return self.base_resolution

        x_density = (self.base_resolution[0] - 1) / base_width
        y_density = (self.base_resolution[1] - 1) / base_height

        adapted_width = adaptive_bounds.width()
        adapted_height = adaptive_bounds.height()

        # Grid resolution must be at least MIN_GRID_RESOLUTION x MIN_GRID_RESOLUTION
        x_res = max(MIN_GRID_RESOLUTION, ceil(adapted_width * x_density) + 1)
        y_res = max(MIN_GRID_RESOLUTION, ceil(adapted_height * y_density) + 1)

        return x_res, y_res



from __future__ import annotations

from typing import TYPE_CHECKING, Protocol

if TYPE_CHECKING:
    from cartographer.core import MacroRegistration, PrinterCartographer
    from cartographer.interfaces.printer import Endstop


class Integrator(Protocol):
    def setup(self) -> None: ...
    def register_cartographer(self, cartographer: PrinterCartographer) -> None: ...
    def register_macro(self, registration: MacroRegistration) -> None: ...
    def register_endstop_pin(self, chip_name: str, pin: str, endstop: Endstop) -> None: ...
    def register_temperature_sensor_factories(self) -> None: ...



from __future__ import annotations

import logging
from abc import ABC
from functools import wraps
from textwrap import dedent
from typing import TYPE_CHECKING, Callable, Protocol, final

from gcode import CommandError, GCodeCommand, GCodeDispatch
from typing_extensions import override

from cartographer.adapters.klipper.endstop import KlipperEndstop, KlipperHomingState
from cartographer.adapters.klipper.homing import KlipperHomingChip
from cartographer.adapters.klipper.logging import setup_console_logger
from cartographer.adapters.klipper.temperature import PrinterTemperatureCoil
from cartographer.adapters.utils import reraise_as
from cartographer.interfaces.printer import Macro, MacroParams, SupportsFallbackMacro
from cartographer.runtime.integrator import Integrator

if TYPE_CHECKING:
    from extras.homing import Homing
    from klippy import Printer
    from stepper import PrinterRail

    from cartographer.adapters.klipper.configuration import KlipperConfiguration
    from cartographer.adapters.klipper.mcu import KlipperCartographerMcu
    from cartographer.adapters.klipper.toolhead import KlipperToolhead
    from cartographer.core import MacroRegistration
    from cartographer.interfaces.printer import Endstop

logger = logging.getLogger(__name__)


class KlipperLikeAdapters(Protocol):
    mcu: KlipperCartographerMcu
    toolhead: KlipperToolhead
    printer: Printer
    config: KlipperConfiguration


class KlipperLikeIntegrator(Integrator, ABC):
    def __init__(self, adapters: KlipperLikeAdapters) -> None:
        self._config: KlipperConfiguration = adapters.config
        self._printer: Printer = adapters.printer
        self._mcu: KlipperCartographerMcu = adapters.mcu
        self._toolhead: KlipperToolhead = adapters.toolhead

        self._gcode: GCodeDispatch = self._printer.lookup_object("gcode")

    @override
    def setup(self) -> None:
        self._printer.register_event_handler("homing:home_rails_end", self._handle_home_rails_end)
        self._configure_macro_logger()

    @override
    def register_endstop_pin(self, chip_name: str, pin: str, endstop: Endstop) -> None:
        mcu_endstop = KlipperEndstop(self._mcu, endstop)
        chip = KlipperHomingChip(mcu_endstop, pin)
        self._printer.lookup_object("pins").register_chip(chip_name, chip)

    @override
    def register_macro(self, registration: MacroRegistration) -> None:
        name = registration.name
        macro = registration.macro
        if isinstance(macro, SupportsFallbackMacro):
            original = self._gcode.register_command(name, None)
            if original:
                macro.set_fallback_macro(FallbackMacroAdapter(name, original))
            else:
                logger.warning("No original macro found to fallback to for '%s'", name)

        self._gcode.register_command(name, _catch_macro_errors(macro.run), desc=macro.description)

    @override
    def register_temperature_sensor_factories(self) -> None:
        self._printer.load_object(self._config.wrapper, "heaters").add_sensor_factory(
            "cartographer_coil", PrinterTemperatureCoil
        )

    @reraise_as(CommandError)
    def _handle_home_rails_end(self, homing: Homing, rails: list[PrinterRail]) -> None:
        homing_state = KlipperHomingState(homing)
        klipper_endstops = [
            es.endstop for rail in rails for es, _ in rail.get_endstops() if isinstance(es, KlipperEndstop)
        ]
        for endstop in klipper_endstops:
            endstop.on_home_end(homing_state)

    def _configure_macro_logger(self) -> None:
        handler = setup_console_logger(self._gcode)
        log_level = logging.DEBUG if self._config.general.verbose else logging.INFO
        handler.setLevel(log_level)


def _catch_macro_errors(func: Callable[[GCodeCommand], None]) -> Callable[[GCodeCommand], None]:
    @wraps(func)
    def wrapper(gcmd: GCodeCommand) -> None:
        try:
            func(gcmd)
        except (RuntimeError, ValueError) as e:
            msg = dedent(str(e)).replace("\n", " ").replace("  ", "\n").strip()
            raise gcmd.error(msg) from e

    return wrapper


@final
class FallbackMacroAdapter(Macro):
    def __init__(self, name: str, handler: Callable[[GCodeCommand], None]) -> None:
        self.name = name
        self.description: str = f"Fallback for {name}"
        self._handler: Callable[[GCodeCommand], None] = handler

    @override
    def run(self, params: MacroParams) -> None:
        assert isinstance(params, GCodeCommand), f"Invalid gcode params type for {self.name}"
        self._handler(params)



from __future__ import annotations

import sys
from typing import TYPE_CHECKING

import pytest

from cartographer.interfaces.printer import HomingState, MacroParams, Mcu, Position, Sample, TemperatureStatus, Toolhead
from cartographer.probe.probe import Probe
from cartographer.probe.scan_mode import ScanMode, ScanModeConfiguration
from cartographer.probe.touch_mode import TouchMode, TouchModeConfiguration
from cartographer.stream import Session
from tests.mocks.config import MockConfiguration
from tests.mocks.params import MockParams
from tests.mocks.task_executor import InlineTaskExecutor

if TYPE_CHECKING:
    from pytest_mock import MockerFixture

    from cartographer.interfaces.configuration import Configuration
    from cartographer.interfaces.multiprocessing import TaskExecutor

collect_ignore: list[str] = []
if sys.version_info < (3, 9):
    # pytest-bdd 8.0.0 requires Python 3.9+
    collect_ignore.append("bdd")


@pytest.fixture
def toolhead(mocker: MockerFixture) -> Toolhead:
    mock = mocker.MagicMock(spec=Toolhead, autospec=True, instance=True)

    def get_position() -> Position:
        return Position(x=10, y=10, z=5)

    def apply_axis_twist_compensation(position: Position) -> Position:
        return position

    def get_extruder_temperature() -> TemperatureStatus:
        return TemperatureStatus(30, 30)

    mock.get_position = get_position
    mock.apply_axis_twist_compensation = apply_axis_twist_compensation
    mock.get_extruder_temperature = get_extruder_temperature
    last_move_time = 0

    def get_last_move_time() -> float:
        nonlocal last_move_time
        last_move_time += 1
        return last_move_time

    mock.get_last_move_time = get_last_move_time

    mock.z_homing_move = mocker.Mock(return_value=0)

    return mock


@pytest.fixture
def session(mocker: MockerFixture) -> Session[Sample]:
    return Session(mocker.Mock(), mocker.Mock())


@pytest.fixture
def mcu(mocker: MockerFixture, session: Session[Sample]) -> Mcu:
    mock = mocker.MagicMock(spec=Mcu, autospec=True, instance=True)
    mock.start_session = mocker.Mock(return_value=session)
    return mock


@pytest.fixture
def params() -> MacroParams:
    return MockParams()


@pytest.fixture
def config() -> Configuration:
    return MockConfiguration()


@pytest.fixture
def scan(mcu: Mcu, toolhead: Toolhead, config: Configuration):
    return ScanMode(mcu, toolhead, ScanModeConfiguration.from_config(config))


@pytest.fixture
def touch(mcu: Mcu, toolhead: Toolhead, config: Configuration):
    return TouchMode(mcu, toolhead, TouchModeConfiguration.from_config(config))


@pytest.fixture
def probe(scan: ScanMode, touch: TouchMode) -> Probe:
    return Probe(scan, touch)


@pytest.fixture
def task_executor() -> TaskExecutor:
    return InlineTaskExecutor()


@pytest.fixture
def homing_state(mocker: MockerFixture) -> HomingState:
    return mocker.Mock(spec=HomingState, autospec=True)



from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
import pytest
from typing_extensions import TypeAlias

from cartographer.macros.bed_mesh.paths.random_path import RandomPathGenerator
from cartographer.macros.bed_mesh.paths.snake_path import SnakePathGenerator
from cartographer.macros.bed_mesh.paths.spiral_path import SpiralPathGenerator

if TYPE_CHECKING:
    from cartographer.macros.bed_mesh.interfaces import PathGenerator, Point


def make_grid(nx: int, ny: int, spacing: float) -> list[Point]:
    return [(x * spacing, y * spacing) for y in range(ny) for x in range(nx)]


GeneratorFixture: TypeAlias = "tuple[str, PathGenerator]"
GridFixture: TypeAlias = "tuple[str, list[Point]]"


@pytest.fixture(
    params=[
        ("Snake X", lambda: SnakePathGenerator(main_direction="x")),
        ("Snake Y", lambda: SnakePathGenerator(main_direction="y")),
        ("Spiral", lambda: SpiralPathGenerator(main_direction="x")),
        ("Random", lambda: RandomPathGenerator(main_direction="x")),
    ]
)
def generator(request: pytest.FixtureRequest):
    name, gen_factory = request.param
    return name, gen_factory()


@pytest.fixture(
    params=[
        ("3x3 grid", make_grid(3, 3, 1.0)),
        ("3x5 grid", make_grid(3, 5, 1.0)),
        ("4x4 grid", make_grid(4, 4, 1.0)),
        ("5x3 grid", make_grid(5, 3, 1.0)),
        ("7x8 grid", make_grid(7, 8, 1.0)),
        ("8x7 grid", make_grid(8, 7, 1.0)),
    ]
)
def grid_points(request: pytest.FixtureRequest):
    return request.param


def test_path_generator_covers_all_points(generator: GeneratorFixture, grid_points: GridFixture):
    gen_name, gen = generator
    grid_name, points = grid_points

    max_dist = 0.2  # maximum allowed miss distance per input point

    path = list(gen.generate_path(points, (0, 100), (0, 100)))
    path_array = np.array(path)

    for _, pt in enumerate(points):
        pt = np.array(pt)
        dists = np.linalg.norm(path_array - pt, axis=1)
        min_dist = np.min(dists)
        assert min_dist <= max_dist, (
            f"{gen_name} did not reach input point {pt.tolist()} (min dist {min_dist:.3f} > {max_dist}) on {grid_name}"
        )

    # Optional: continuity check
    max_step = 10.0
    for p0, p1 in zip(path, path[1:]):
        dist = np.linalg.norm(np.array(p1) - np.array(p0))
        assert dist <= max_step, f"{gen_name} discontinuity {dist:.2f} on {grid_name}"



from __future__ import annotations

import logging
from abc import ABC, abstractmethod
from dataclasses import replace
from typing import TYPE_CHECKING, final

from typing_extensions import override

from cartographer.interfaces.configuration import Configuration, TouchModelConfiguration
from cartographer.interfaces.printer import Macro, MacroParams, Mcu
from cartographer.lib.statistics import compute_mad
from cartographer.macros.utils import get_choice
from cartographer.probe.touch_mode import MAD_TOLERANCE, TouchMode, TouchModeConfiguration

if TYPE_CHECKING:
    from cartographer.interfaces.printer import Toolhead
    from cartographer.probe.probe import Probe


logger = logging.getLogger(__name__)


MIN_ALLOWED_STEP = 75
MAX_ALLOWED_STEP = 500
DEFAULT_TOUCH_MODEL_NAME = "default"
DEFAULT_Z_OFFSET = -0.05


class CalibrationStrategy(ABC):
    @abstractmethod
    def compute_score(self, samples: list[float]) -> float: ...
    @abstractmethod
    def is_acceptable(self, score: float) -> bool: ...
    @abstractmethod
    def should_stop_early(self, samples: list[float]) -> bool: ...
    @abstractmethod
    def compute_step_increase(self, current_threshold: int, score: float) -> int: ...


@final
class DefaultCalibrationStrategy(CalibrationStrategy):
    """Standard MAD-based calibration strategy."""

    @override
    def compute_score(self, samples: list[float]) -> float:
        return compute_mad(samples)

    @override
    def is_acceptable(self, score: float) -> bool:
        return score <= MAD_TOLERANCE

    @override
    def should_stop_early(self, samples: list[float]) -> bool:
        if len(samples) < 3:
            return False
        return self.compute_score(samples) > 5 * MAD_TOLERANCE

    @override
    def compute_step_increase(self, current_threshold: int, score: float) -> int:
        allowed_step = (current_threshold * 0.06) ** 1.15
        allowed_max_step = max(MIN_ALLOWED_STEP, min(MAX_ALLOWED_STEP, allowed_step))
        ratio = max(1.0, score / MAD_TOLERANCE)
        raw_step = allowed_max_step * (1 - (1 / ratio))
        return int(round(min(allowed_max_step, max(MIN_ALLOWED_STEP, raw_step))))


@final
class AggressiveCalibrationStrategy(CalibrationStrategy):
    """Faster calibration with larger threshold steps."""

    @override
    def compute_score(self, samples: list[float]) -> float:
        return compute_mad(samples)

    @override
    def is_acceptable(self, score: float) -> bool:
        return score <= MAD_TOLERANCE * 1.2  # Slightly more lenient

    @override
    def should_stop_early(self, samples: list[float]) -> bool:
        if len(samples) < 3:
            return False
        return self.compute_score(samples) > 10 * MAD_TOLERANCE  # Only stop for extreme cases

    @override
    def compute_step_increase(self, current_threshold: int, score: float) -> int:
        allowed_step = (current_threshold * 0.12) ** 1.15
        allowed_max_step = max(MIN_ALLOWED_STEP, min(MAX_ALLOWED_STEP, allowed_step))
        ratio = max(1.0, score / MAD_TOLERANCE)
        raw_step = allowed_max_step * (1 - (1 / ratio))
        return int(round(min(allowed_max_step, max(MIN_ALLOWED_STEP, raw_step))))


@final
class PrecisionCalibrationStrategy(CalibrationStrategy):
    """Robust yet tolerant precision strategy."""

    RANGE_CAP = MAD_TOLERANCE * 2.0  # Prevent one outlier from dominating

    @override
    def compute_score(self, samples: list[float]) -> float:
        mad = compute_mad(samples)
        sample_range = max(samples) - min(samples)

        # Cap the range contribution to prevent overreaction
        capped_range = min(sample_range / 2, self.RANGE_CAP)

        # Weighted score — mostly MAD, with light influence from range
        return 0.75 * mad + 0.25 * capped_range

    @override
    def is_acceptable(self, score: float) -> bool:
        return score <= MAD_TOLERANCE * 1.0  # Slightly relaxed

    @override
    def should_stop_early(self, samples: list[float]) -> bool:
        if len(samples) < 5:
            return False
        return self.compute_score(samples) > MAD_TOLERANCE * 3.0

    @override
    def compute_step_increase(self, current_threshold: int, score: float) -> int:
        allowed_step = (current_threshold * 0.05) ** 1.15
        allowed_max_step = max(MIN_ALLOWED_STEP, min(MAX_ALLOWED_STEP, allowed_step))
        ratio = max(1.0, score / MAD_TOLERANCE)
        raw_step = allowed_max_step * (1 - (1 / ratio))
        return int(round(min(allowed_max_step, max(MIN_ALLOWED_STEP, raw_step))))


STRATEGY_MAP = {
    "default": DefaultCalibrationStrategy,
    "aggressive": AggressiveCalibrationStrategy,
    "precision": PrecisionCalibrationStrategy,
}


@final
class TouchCalibrateMacro(Macro):
    description = "Run the touch calibration"

    def __init__(self, probe: Probe, mcu: Mcu, toolhead: Toolhead, config: Configuration) -> None:
        self._probe = probe
        self._mcu = mcu
        self._toolhead = toolhead
        self._config = config

    @override
    def run(self, params: MacroParams) -> None:
        name = params.get("MODEL_NAME", DEFAULT_TOUCH_MODEL_NAME)
        speed = params.get_int("SPEED", default=3, minval=1, maxval=5)
        threshold_start = params.get_int("START", default=500, minval=100)
        threshold_max = params.get_int("MAX", default=3000, minval=threshold_start)
        strategy_type = get_choice(params, "STRATEGY", default="default", choices=STRATEGY_MAP.keys())
        strategy = STRATEGY_MAP[strategy_type]()

        if not self._toolhead.is_homed("x") or not self._toolhead.is_homed("y"):
            msg = "Must home x and y before calibration"
            raise RuntimeError(msg)

        self._toolhead.move(
            x=self._config.bed_mesh.zero_reference_position[0],
            y=self._config.bed_mesh.zero_reference_position[1],
            speed=self._config.general.travel_speed,
        )
        self._toolhead.wait_moves()

        logger.info(
            "Starting %s touch calibration at speed %d, threshold %d-%d",
            strategy_type,
            speed,
            threshold_start,
            threshold_max,
        )

        calibration_mode = CalibrationTouchMode(
            self._mcu,
            self._toolhead,
            TouchModeConfiguration.from_config(self._config),
            threshold=threshold_start,
            speed=speed,
            strategy=strategy,
        )

        forced_z = False
        try:
            if not self._toolhead.is_homed("z"):
                forced_z = True
                _, z_max = self._toolhead.get_axis_limits("z")
                self._toolhead.set_z_position(z=z_max - 10)

            threshold = self._find_acceptable_threshold(calibration_mode, threshold_start, threshold_max)
        finally:
            if forced_z:
                self._toolhead.clear_z_homing_state()

        if threshold is None:
            logger.info(
                """
                Failed to calibrate with %s strategy (thresholds %d-%d).
                Try increasing MAX or changing strategy.
                CARTOGRAPHER_TOUCH_CALIBRATE START=%d MAX=%d
                """,
                strategy_type,
                threshold_start,
                threshold_max,
                threshold_max,
                threshold_max + 2000,
            )
            return

        logger.info(
            "Successfully calibrated with %s strategy (threshold %d, speed %.1f)", strategy_type, threshold, speed
        )
        model = TouchModelConfiguration(name, threshold, speed, DEFAULT_Z_OFFSET)
        self._config.save_touch_model(model)
        logger.info(
            """
            Touch model %s has been saved for the current session.
            The SAVE_CONFIG command will update the printer config file and restart the printer.
            """,
            name,
        )

    def _find_acceptable_threshold(
        self, calibration_mode: CalibrationTouchMode, threshold_start: int, threshold_max: int
    ) -> int | None:
        current_threshold = threshold_start
        strategy = calibration_mode.strategy

        while current_threshold <= threshold_max:
            samples = calibration_mode.collect_samples(current_threshold)
            score = strategy.compute_score(samples)

            logger.info(
                "Threshold %d score: %.6f",
                current_threshold,
                score,
            )
            logger.debug("Samples: %s", ", ".join(f"{s:.6f}" for s in samples))

            if strategy.is_acceptable(score):
                logger.info("Threshold %d accepted (score %.6f)", current_threshold, score)
                return current_threshold

            step = strategy.compute_step_increase(current_threshold, score)
            current_threshold += step
            logger.info("Next threshold: %d (+%d)", current_threshold, step)

        return None


@final
class CalibrationTouchMode(TouchMode):
    def __init__(
        self,
        mcu: Mcu,
        toolhead: Toolhead,
        config: TouchModeConfiguration,
        *,
        threshold: int,
        speed: float,
        strategy: CalibrationStrategy,
    ) -> None:
        model = TouchModelConfiguration("calibration", threshold, speed, 0)
        super().__init__(mcu, toolhead, replace(config, models={"calibration": model}))
        self.load_model("calibration")
        self.strategy = strategy

    def set_threshold(self, threshold: int) -> None:
        self._models["calibration"] = replace(self._models["calibration"], threshold=threshold)
        self.load_model("calibration")

    def collect_samples(self, threshold: int) -> list[float]:
        samples: list[float] = []
        self.set_threshold(threshold)

        for _ in range(self._config.samples * 3):
            pos = self._perform_single_probe()
            samples.append(pos)

            if self.strategy.should_stop_early(samples):
                break

        return sorted(samples)



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, cast, final

from gcode import CommandError
from mcu import MCU_endstop
from typing_extensions import override

from cartographer.adapters.utils import reraise_as
from cartographer.interfaces.printer import HomingAxis, HomingState

if TYPE_CHECKING:
    from extras.homing import Homing
    from mcu import MCU
    from reactor import ReactorCompletion
    from stepper import MCU_stepper

    from cartographer.adapters.klipper.mcu import KlipperCartographerMcu
    from cartographer.interfaces.printer import Endstop

logger = logging.getLogger(__name__)

axis_mapping: dict[HomingAxis, int] = {
    "x": 0,
    "y": 1,
    "z": 2,
}


def axis_to_index(axis: HomingAxis) -> int:
    return axis_mapping[axis]


@final
class KlipperHomingState(HomingState):
    def __init__(self, homing: Homing) -> None:
        self.homing = homing

    @override
    def is_homing_z(self) -> bool:
        return axis_to_index("z") in self.homing.get_axes()

    @override
    def set_z_homed_position(self, position: float) -> None:
        logger.debug("Setting homed distance for z to %.3f", position)
        self.homing.set_homed_position([None, None, position])


@final
class KlipperEndstop(MCU_endstop):
    def __init__(self, mcu: KlipperCartographerMcu, endstop: Endstop):
        self.mcu = mcu
        self.endstop = endstop

    @override
    def get_mcu(self) -> MCU:
        return self.mcu.klipper_mcu

    @override
    def add_stepper(self, stepper: MCU_stepper) -> None:
        return self.mcu.dispatch.add_stepper(stepper)

    @override
    def get_steppers(self) -> list[MCU_stepper]:
        return self.mcu.dispatch.get_steppers()

    @override
    @reraise_as(CommandError)
    def home_start(
        self,
        print_time: float,
        sample_time: float,
        sample_count: int,
        rest_time: float,
        triggered: bool = True,
    ) -> ReactorCompletion:
        del sample_time, sample_count, rest_time, triggered
        return cast("ReactorCompletion", self.endstop.home_start(print_time))

    @override
    @reraise_as(CommandError)
    def home_wait(self, home_end_time: float) -> float:
        return self.endstop.home_wait(home_end_time)

    @override
    @reraise_as(CommandError)
    def query_endstop(self, print_time: float) -> int:
        return 1 if self.endstop.query_is_triggered(print_time) else 0

    @override
    def get_position_endstop(self) -> float:
        return self.endstop.get_endstop_position()



from typing import Callable, Literal, Protocol, TypedDict

from configfile import ConfigWrapper
from gcode import GCodeCommand
from mcu import MCU_endstop
from typing_extensions import override

HINT_TIMEOUT: str

type _Pos = list[float]

def run_single_probe(probe: PrinterProbe, gcmd: GCodeCommand) -> None: ...
def calc_probe_z_average(positions: list[_Pos], method: Literal["average", "median"] = "average") -> _Pos: ...

class _ProbeStatus(TypedDict):
    name: str
    last_query: bool | int
    last_z_result: float

class ProbeCommandHelper:
    def __init__(
        self,
        config: ConfigWrapper,
        probe: object,
        query_endstop: Callable[[float], int] | None = None,
    ) -> None: ...
    def get_status(self, eventtime: float) -> _ProbeStatus: ...

class HomingViaProbeHelper:
    def __init__(self, config: ConfigWrapper, mcu_probe: ProbeEndstopWrapper) -> None: ...

class ProbeOffsetsHelper:
    def __init__(self, config: ConfigWrapper) -> None: ...
    def get_offsets(self) -> tuple[float, float, float]: ...

class _ProbeSession(Protocol):
    def run_probe(self, gcmd: GCodeCommand) -> None: ...
    def end_probe_session(self) -> None: ...
    def pull_probed_results(self) -> list[_Pos]: ...

class ProbeSessionHelper(_ProbeSession):
    def __init__(self, config: ConfigWrapper, mcu_probe: ProbeEndstopWrapper) -> None: ...
    def start_probe_session(self, gcmd: GCodeCommand) -> _ProbeSession: ...
    def get_probe_params(self, gcmd: GCodeCommand | None = None) -> dict[str, str | int | float]: ...
    @override
    def run_probe(self, gcmd: GCodeCommand) -> None: ...
    @override
    def pull_probed_results(self) -> list[_Pos]: ...
    @override
    def end_probe_session(self) -> None: ...

class ProbeEndstopWrapper(MCU_endstop, Protocol):
    def multi_probe_begin(self) -> None: ...
    def multi_probe_end(self) -> None: ...
    def probing_move(self, pos: _Pos, speed: float) -> _Pos: ...
    def probe_prepare(self, hmove: float) -> None: ...
    def probe_finish(self, hmove: float) -> None: ...
    @override
    def get_position_endstop(self) -> float: ...

class PrinterProbe(Protocol):
    def get_offsets(self) -> tuple[float, float, float]: ...
    def get_status(self, eventtime: float) -> dict[str, str | int | float]: ...
    # def get_probe_params(self, gcmd: GCodeCommand | None = None) -> dict[str, str | int | float]: ...
    # def start_probe_session(self, gcmd: GCodeCommand) -> _ProbeSession: ...



from __future__ import annotations

from typing import overload

from typing_extensions import override

from cartographer.interfaces.printer import MacroParams


class MockParams(MacroParams):
    def __init__(self) -> None:
        self.params: dict[str, str] = {}

    @overload
    def get(self, name: str, default: str = ...) -> str: ...
    @overload
    def get(self, name: str, default: None) -> str | None: ...

    @override
    def get(self, name: str, default: str | None = ...) -> str | None:
        opt = self.params.get(name, default)
        return str(opt) if opt is not None else None

    @overload
    def get_float(
        self, name: str, default: float = ..., *, above: float = ..., minval: float = ..., maxval: float = ...
    ) -> float: ...
    @overload
    def get_float(
        self, name: str, default: None, *, above: float = ..., minval: float = ..., maxval: float = ...
    ) -> float | None: ...

    @override
    def get_float(
        self, name: str, default: float | None = ..., *, above: float = ..., minval: float = ..., maxval: float = ...
    ) -> float | None:
        opt = self.params.get(name, default)
        return float(opt) if opt is not None else None

    @override
    def get_int(self, name: str, default: int = ..., *, minval: int = ..., maxval: int = ...) -> int:
        return int(self.params.get(name, default))



from __future__ import annotations

import logging
from dataclasses import dataclass, replace
from itertools import chain
from math import isfinite
from typing import TYPE_CHECKING, Literal, final

from typing_extensions import override

from cartographer.interfaces.printer import Macro, MacroParams, Position, Sample, SupportsFallbackMacro, Toolhead
from cartographer.lib.log import log_duration
from cartographer.macros.bed_mesh.helpers import (
    AdaptiveMeshCalculator,
    CoordinateTransformer,
    GridPointResult,
    MeshBounds,
    MeshGrid,
    SampleProcessor,
)
from cartographer.macros.bed_mesh.paths.alternating_snake import AlternatingSnakePathGenerator
from cartographer.macros.bed_mesh.paths.random_path import RandomPathGenerator
from cartographer.macros.bed_mesh.paths.snake_path import SnakePathGenerator
from cartographer.macros.bed_mesh.paths.spiral_path import SpiralPathGenerator
from cartographer.macros.utils import get_choice, get_float_tuple, get_int_tuple

if TYPE_CHECKING:
    from cartographer.interfaces.configuration import Configuration
    from cartographer.interfaces.multiprocessing import TaskExecutor
    from cartographer.macros.bed_mesh.interfaces import BedMeshAdapter, PathGenerator, Point
    from cartographer.probe import Probe

logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class BedMeshCalibrateConfiguration:
    mesh_min: tuple[float, float]
    mesh_max: tuple[float, float]
    probe_count: tuple[int, int]
    speed: float
    adaptive_margin: float
    zero_reference_position: Point

    runs: int
    direction: Literal["x", "y"]
    height: float
    path: Literal["snake", "alternating_snake", "spiral", "random"]

    @staticmethod
    def from_config(config: Configuration):
        return BedMeshCalibrateConfiguration(
            mesh_min=config.bed_mesh.mesh_min,
            mesh_max=config.bed_mesh.mesh_max,
            probe_count=config.bed_mesh.probe_count,
            speed=config.bed_mesh.speed,
            adaptive_margin=config.bed_mesh.adaptive_margin,
            zero_reference_position=config.bed_mesh.zero_reference_position,
            runs=config.scan.mesh_runs,
            direction=config.scan.mesh_direction,
            height=config.scan.mesh_height,
            path=config.scan.mesh_path,
        )


_directions: list[Literal["x", "y"]] = ["x", "y"]

PATH_GENERATOR_MAP = {
    "snake": SnakePathGenerator,
    "alternating_snake": AlternatingSnakePathGenerator,
    "spiral": SpiralPathGenerator,
    "random": RandomPathGenerator,
}


@dataclass
class MeshScanParams:
    mesh_bounds: MeshBounds
    resolution: tuple[int, int]
    speed: float
    height: float
    runs: int
    adaptive: bool
    adaptive_margin: float
    profile: str | None
    path_generator: PathGenerator

    @classmethod
    def from_macro_params(
        cls, params: MacroParams, config: BedMeshCalibrateConfiguration, adapter: BedMeshAdapter
    ) -> MeshScanParams:
        """Create parameters from macro input and configuration."""
        base_bounds = MeshBounds(
            get_float_tuple(params, "MESH_MIN", default=config.mesh_min),
            get_float_tuple(params, "MESH_MAX", default=config.mesh_max),
        )
        base_resolution = get_int_tuple(params, "PROBE_COUNT", default=config.probe_count)

        adaptive = params.get_int("ADAPTIVE", default=0) != 0
        adaptive_margin = params.get_float("ADAPTIVE_MARGIN", config.adaptive_margin, minval=0)

        # Calculate actual bounds and resolution
        if adaptive:
            calculator = AdaptiveMeshCalculator(base_bounds, base_resolution)
            object_points = list(chain.from_iterable(adapter.get_objects()))
            mesh_bounds = calculator.calculate_adaptive_bounds(object_points, adaptive_margin)
            resolution = calculator.calculate_adaptive_resolution(mesh_bounds)
            profile = None  # Adaptive meshes don't use profiles
        else:
            mesh_bounds = base_bounds
            resolution = base_resolution
            profile = params.get("PROFILE", default="default")

        # Create path generator
        direction: Literal["x", "y"] = get_choice(params, "DIRECTION", _directions, default=config.direction)
        path_type = get_choice(params, "PATH", default=config.path, choices=PATH_GENERATOR_MAP.keys())
        path_generator = PATH_GENERATOR_MAP[path_type](direction)

        return cls(
            mesh_bounds=mesh_bounds,
            resolution=resolution,
            speed=params.get_float("SPEED", default=config.speed, minval=50),
            height=params.get_float("HEIGHT", default=config.height, minval=0.5, maxval=5),
            runs=params.get_int("RUNS", default=config.runs, minval=1),
            adaptive=adaptive,
            adaptive_margin=adaptive_margin,
            profile=profile,
            path_generator=path_generator,
        )


@final
class BedMeshCalibrateMacro(Macro, SupportsFallbackMacro):
    description = "Gather samples across the bed to calibrate the bed mesh."

    def __init__(
        self,
        probe: Probe,
        toolhead: Toolhead,
        adapter: BedMeshAdapter,
        task_executor: TaskExecutor,
        config: BedMeshCalibrateConfiguration,
    ):
        self.probe = probe
        self.toolhead = toolhead
        self.adapter = adapter
        self.task_executor = task_executor
        self.config = config
        self.coordinate_transformer = CoordinateTransformer(probe.scan.offset)
        self._fallback: Macro | None = None

    @override
    def set_fallback_macro(self, macro: Macro) -> None:
        self._fallback = macro

    @override
    def run(self, params: MacroParams) -> None:
        """Main entry point for bed mesh calibration."""
        # Handle fallback for non-scan methods
        method = params.get("METHOD", "scan")
        if method.lower() != "scan":
            if self._fallback is None:
                msg = f"Bed mesh calibration method '{method}' not supported"
                raise RuntimeError(msg)
            return self._fallback.run(params)

        # Parse parameters and validate
        scan_params = MeshScanParams.from_macro_params(params, self.config, self.adapter)
        self._validate_zero_reference(scan_params.mesh_bounds)

        # Create mesh grid and processors
        grid = MeshGrid(
            scan_params.mesh_bounds.min_point,
            scan_params.mesh_bounds.max_point,
            scan_params.resolution[0],
            scan_params.resolution[1],
        )
        # Generate path and collect samples
        path = self._generate_path(grid, scan_params)
        samples = self._collect_samples(path, scan_params)

        # Process samples and create mesh
        positions = self.task_executor.run(self._process_samples_to_positions, grid, samples, scan_params.height)

        # Apply mesh to adapter
        self.adapter.clear_mesh()
        self.adapter.apply_mesh(positions, scan_params.profile)

    def _validate_zero_reference(self, mesh_bounds: MeshBounds) -> None:
        """Validate that zero reference position is within mesh bounds."""
        zero_ref = self.config.zero_reference_position

        if not (
            mesh_bounds.min_point[0] <= zero_ref[0] <= mesh_bounds.max_point[0]
            and mesh_bounds.min_point[1] <= zero_ref[1] <= mesh_bounds.max_point[1]
        ):
            zrp = f"({zero_ref[0]:.2f},{zero_ref[1]:.2f})"
            bounds_str = (
                f"({mesh_bounds.min_point[0]:.2f},{mesh_bounds.min_point[1]:.2f}) - "
                f"({mesh_bounds.max_point[0]:.2f},{mesh_bounds.max_point[1]:.2f})"
            )
            msg = f"Zero reference position {zrp} must be within mesh bounds {bounds_str}"
            raise RuntimeError(msg)

    def _generate_path(self, grid: MeshGrid, params: MeshScanParams) -> list[Point]:
        """Generate scanning path from grid points."""
        mesh_points = grid.generate_points()
        x_axis_limits = self.toolhead.get_axis_limits("x")
        y_axis_limits = self.toolhead.get_axis_limits("y")

        return list(params.path_generator.generate_path(mesh_points, x_axis_limits, y_axis_limits))

    @log_duration("Collecting samples along the scanning path")
    def _collect_samples(self, path: list[Point], params: MeshScanParams) -> list[Sample]:
        """Collect samples by following the scanning path."""
        # Move to starting position
        self.toolhead.move(z=params.height, speed=5)
        self._move_probe_to_point(path[0], params.speed)
        self.toolhead.wait_moves()

        # Execute scan
        with self.probe.scan.start_session() as session:
            session.wait_for(lambda samples: len(samples) >= 10)

            for run_index in range(params.runs):
                sequence = path if run_index % 2 == 0 else reversed(path)
                for point in sequence:
                    self._move_probe_to_point(point, params.speed)

                self.toolhead.dwell(0.250)
                self.toolhead.wait_moves()

            # Wait for final samples
            move_time = self.toolhead.get_last_move_time()
            session.wait_for(lambda samples: samples[-1].time >= move_time)
            count = len(session.items)
            session.wait_for(lambda samples: len(samples) >= count + 10)

        samples = session.get_items()
        logger.debug("Collected %d samples across %d runs", len(samples), params.runs)
        return [self._transform_sample(s) for s in samples]

    def _move_probe_to_point(self, point: Point, speed: float) -> None:
        """Move probe to specified point (converts to nozzle coordinates)."""
        x, y = self.coordinate_transformer.probe_to_nozzle(point)
        self.toolhead.move(x=float(x), y=float(y), speed=speed)

    def _transform_sample(self, sample: Sample) -> Sample:
        """Transform sample to probe coordinates."""
        if sample.position is None:
            return sample

        probe_position = self.coordinate_transformer.nozzle_to_probe((sample.position.x, sample.position.y))
        return replace(
            sample, position=Position(x=float(probe_position[0]), y=float(probe_position[1]), z=sample.position.z)
        )

    @log_duration("Processing samples into final mesh positions")
    def _process_samples_to_positions(self, grid: MeshGrid, samples: list[Sample], height: float) -> list[Position]:
        """Process samples into final mesh positions."""
        sample_processor = SampleProcessor(grid)

        # Assign samples to grid points
        results = sample_processor.assign_samples_to_grid(samples, self.probe.scan.calculate_sample_distance)

        # Convert results to positions
        positions = self._results_to_positions(results, height)

        # Normalize to zero reference
        return self.coordinate_transformer.normalize_to_zero_reference(positions, self.config.zero_reference_position)

    def _results_to_positions(self, results: list[GridPointResult], height: float) -> list[Position]:
        """Convert grid results to Position objects."""
        positions: list[Position] = []

        for result in results:
            rx, ry = result.point
            if not isfinite(result.z):
                msg = f"Grid point ({rx:.2f},{ry:.2f}) has no valid samples"
                raise RuntimeError(msg)

            # Calculate compensated height
            z = height - result.z
            nx, ny = self.coordinate_transformer.probe_to_nozzle(result.point)
            compensated = self.toolhead.apply_axis_twist_compensation(Position(x=float(nx), y=float(ny), z=z))

            # Convert back to probe coordinates
            positions.append(Position(x=float(rx), y=float(ry), z=compensated.z))

        return positions



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Callable, Generic, Protocol, TypeVar

T = TypeVar("T")


class Condition(Protocol):
    def notify_all(self) -> None:
        """Wakes all threads waiting on this condition."""

    def wait_for(self, predicate: Callable[[], bool]) -> None:
        """Wait until a condition evaluates to True."""
        ...


class Session(Generic[T]):
    def __init__(
        self,
        stream: Stream[T],
        condition: Condition,
        start_condition: Callable[[T], bool] | None = None,
    ):
        self.stream: Stream[T] = stream
        self.items: list[T] = []
        self.start_condition: Callable[[T], bool] | None = start_condition
        self._condition: Condition = condition

    def add_item(self, item: T):
        """Adds an item to the session only after the start condition is met."""
        if self.start_condition is not None:
            if not self.start_condition(item):
                return
            self.start_condition = None

        self.items.append(item)
        self._condition.notify_all()

    def wait_for(self, condition: Callable[[list[T]], bool]):
        """Waits until the given condition function returns True."""
        self._condition.wait_for(lambda: condition(self.items))

    def get_items(self) -> list[T]:
        """Returns collected items after session ends."""
        return self.items

    def __enter__(self):
        return self  # Allows using `with session:`

    def __exit__(
        self,
        exc_type: object,
        exc_val: object,
        exc_tb: object,
    ):
        """Automatically remove session from stream when exiting the with-block."""
        self.stream.end_session(self)


class Stream(ABC, Generic[T]):
    def __init__(self, smoothing_fn: Callable[[T], T] | None = None):
        """Initializes a stream with optional smoothing function."""
        self.smoothing_fn: Callable[[T], T] | None = smoothing_fn
        self.sessions: set[Session[T]] = set()
        self.callbacks: set[Callable[[T], None]] = set()

    @abstractmethod
    def condition(self) -> Condition: ...

    def start_session(self, start_condition: Callable[[T], bool] | None = None) -> Session[T]:
        """Starts a session with an optional start condition.

        All items before the start condition will be skipped.
        """
        session = Session(self, self.condition(), start_condition)
        self.sessions.add(session)
        return session

    def end_session(self, session: Session[T]):
        """Ends a session and removes it from active sessions."""
        self.sessions.discard(session)

    def register_callback(self, callback: Callable[[T], None]):
        """Registers a callback to the stream."""
        self.callbacks.add(callback)

    def unregister_callback(self, callback: Callable[[T], None]):
        """Removes the callback from the stream."""
        self.callbacks.discard(callback)

    def add_item(self, item: T):
        """Pushes the item to all active sessions."""

        if self.smoothing_fn is not None:
            item = self.smoothing_fn(item)

        for session in self.sessions:
            session.add_item(item)

        for callback in self.callbacks:
            callback(item)



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, final

from cartographer.adapters.klipper.axis_twist_compensation import KlipperAxisTwistCompensationHelper
from cartographer.adapters.klipper.bed_mesh import KlipperBedMesh
from cartographer.adapters.klipper.configuration import KlipperConfiguration
from cartographer.adapters.klipper.mcu import KlipperCartographerMcu
from cartographer.adapters.klipper.task_executor import KlipperMultiprocessingExecutor
from cartographer.adapters.klipper.toolhead import KlipperToolhead
from cartographer.runtime.adapters import Adapters

if TYPE_CHECKING:
    from configfile import ConfigWrapper as KlipperConfigWrapper


logger = logging.getLogger(__name__)


@final
class KlipperAdapters(Adapters):
    def __init__(self, config: KlipperConfigWrapper) -> None:
        self.printer = config.get_printer()

        self.config = KlipperConfiguration(config)
        self.mcu = KlipperCartographerMcu(config)
        self.task_executor = KlipperMultiprocessingExecutor(self.printer.get_reactor())

        self.toolhead = KlipperToolhead(config, self.mcu)
        self.bed_mesh = KlipperBedMesh(config)

        self.axis_twist_compensation = None
        if config.has_section("axis_twist_compensation"):
            self.axis_twist_compensation = KlipperAxisTwistCompensationHelper(config)



from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from pytest_bdd import given, parsers, scenarios, then, when

from cartographer.interfaces.printer import HomingState
from cartographer.macros.touch import TouchHomeMacro

if TYPE_CHECKING:
    from pytest import LogCaptureFixture
    from pytest_mock import MockerFixture

    from cartographer.interfaces.configuration import Configuration
    from cartographer.interfaces.printer import MacroParams, Toolhead
    from cartographer.probe import Probe
    from cartographer.probe.touch_mode import TouchMode
    from tests.bdd.helpers.context import Context


scenarios("../../features/z_offset.feature")


@given(parsers.parse("I have baby stepped the nozzle {offset:g}mm up"))
def given_baby_step_up(toolhead: Toolhead, offset: float):
    toolhead.get_gcode_z_offset = lambda: offset


@given(parsers.parse("I have baby stepped the nozzle {offset:g}mm down"))
def given_baby_step_down(toolhead: Toolhead, offset: float):
    toolhead.get_gcode_z_offset = lambda: -offset


@given("I ran G28")
def given_g28(mocker: MockerFixture, probe: Probe):
    homing_state = mocker.Mock(spec=HomingState, autospec=True)
    probe.scan.on_home_end(homing_state)


@given("I ran TOUCH_HOME")
def given_touch_home(touch: TouchMode, toolhead: Toolhead, params: MacroParams):
    macro = TouchHomeMacro(touch, toolhead, (10, 10))
    macro.run(params)
    # homing_state = mocker.Mock(spec=HomingState, autospec=True)
    # probe.touch.on_home_end(homing_state)


@when("I run the Z_OFFSET_APPLY_PROBE macro")
def when_run_probe_accuracy_macro(
    params: MacroParams,
    caplog: LogCaptureFixture,
    probe: Probe,
    toolhead: Toolhead,
    config: Configuration,
    context: Context,
):
    from cartographer.macros.probe import ZOffsetApplyProbeMacro

    macro = ZOffsetApplyProbeMacro(probe, toolhead, config)
    with caplog.at_level(logging.INFO):
        try:
            macro.run(params)
        except Exception as e:
            context.error = e


@then(parsers.parse("it should set scan z-offset to {offset:g}"))
def then_update_scan_z_offset(config: Configuration, offset: str):
    assert config.scan.models["default"].z_offset == float(offset)


@then(parsers.parse("it should set touch z-offset to {offset:g}"))
def then_update_touch_z_offset(config: Configuration, offset: str):
    assert config.touch.models["default"].z_offset == float(offset)



from __future__ import annotations

import logging
import re
from textwrap import dedent
from typing import Protocol

from typing_extensions import override

module_name = __name__.split(".")[0]

root_logger = logging.getLogger(module_name)


class Console(Protocol):
    def respond_raw(self, msg: str) -> None: ...


def setup_console_logger(console: Console) -> logging.Handler:
    console_handler = GCodeConsoleHandler(console)
    console_handler.setFormatter(GCodeConsoleFormatter())
    console_handler.addFilter(GCodeConsoleFilter())
    root_logger.addHandler(console_handler)

    root_logger.setLevel(logging.DEBUG)  # To allow all messages to be handled by the console handler

    return console_handler


MACRO_PATTERN = re.compile(r"\b([A-Z_]+(?:\s+[A-Z0-9_]+=.*?)*)(?=\s|$)")


def format_macro(macro: str) -> str:
    return f'<a class="command">{macro}</a>'


class GCodeConsoleFormatter(logging.Formatter):
    def __init__(self) -> None:
        super().__init__("%(message)s")

    @override
    def format(self, record: logging.LogRecord) -> str:
        prefix = "!! " if record.levelno >= logging.ERROR else ""
        message = dedent(super().format(record)).replace("\n", " ").replace("  ", "\n").strip()
        return prefix + MACRO_PATTERN.sub(lambda m: format_macro(m.group(0)), message)


class GCodeConsoleFilter(logging.Filter):
    def __init__(self) -> None:
        super().__init__("%(message)s")

    @override
    def filter(self, record: logging.LogRecord) -> bool:
        return "klipper.mcu" not in record.name or record.levelno >= logging.WARNING


class GCodeConsoleHandler(logging.Handler):
    def __init__(self, console: Console) -> None:
        self.console: Console = console
        super().__init__()

    @override
    def emit(self, record: logging.LogRecord) -> None:
        try:
            log_entry = self.format(record)
            self.console.respond_raw(f"{log_entry}\n")

        except Exception:
            self.handleError(record)



from __future__ import annotations

from typing import TYPE_CHECKING, final

from gcode import CommandError, GCodeCommand

from cartographer.adapters.utils import reraise_as

if TYPE_CHECKING:
    from cartographer.adapters.klipper.toolhead import KlipperToolhead
    from cartographer.interfaces.printer import ProbeMode
    from cartographer.macros.probe import ProbeMacro, QueryProbeMacro


@final
class KalicoCartographerProbe:
    def __init__(
        self,
        toolhead: KlipperToolhead,
        probe: ProbeMode,
        probe_macro: ProbeMacro,
        query_probe_macro: QueryProbeMacro,
    ) -> None:
        self.probe = probe
        self.probe_macro = probe_macro
        self.query_probe_macro = query_probe_macro
        self.toolhead = toolhead

        self.lift_speed = 5.0
        self.sample_count = 1
        self.samples_tolerance = 0.1
        self.samples_retries = 0

    def get_offsets(self) -> tuple[float, float, float]:
        return self.probe.offset.as_tuple()

    def get_status(self, eventtime: float):
        del eventtime
        return {
            "name": "cartographer",
            "last_query": 1 if self.query_probe_macro.last_triggered else 0,
            "last_z_result": self.probe_macro.last_trigger_position or 0,
        }

    def get_lift_speed(self, gcmd: GCodeCommand | None = None):
        if gcmd is not None:
            return gcmd.get_float("LIFT_SPEED", self.lift_speed, above=0.0)
        return self.lift_speed

    @reraise_as(CommandError)
    def run_probe(self, gcmd: GCodeCommand) -> list[float]:
        del gcmd
        pos = self.toolhead.get_position()
        trigger_pos = self.probe.perform_probe()
        return [pos.x, pos.y, trigger_pos]

    def multi_probe_begin(self):
        pass

    def multi_probe_end(self):
        pass



# https://github.com/Klipper3d/klipper/blob/master/klippy/toolhead.py
from typing import TypedDict

import gcode

from kinematics import Kinematics
from kinematics import Status as KinematicsStatus
from kinematics.extruder import Extruder

class _Status(KinematicsStatus, TypedDict):
    print_time: float
    stalls: int
    estimated_print_time: float
    extruder: str
    position: gcode.Coord
    max_velocity: float
    max_accel: float
    minimum_cruise_ratio: float
    square_corner_velocity: float

type _Pos = list[float]

class ToolHead:
    Coord: type[gcode.Coord]
    def get_kinematics(self) -> Kinematics: ...
    def get_extruder(self) -> Extruder: ...
    def get_status(self, eventtime: float) -> _Status: ...
    def get_position(self) -> _Pos: ...
    def set_position(
        self, newpos: _Pos, homing_axes: str | tuple[int, ...] = ...
    ) -> None: ...  # TODO: Kalico and old klippy takes tuple[int,..]
    def move(self, newpos: _Pos, speed: float) -> None: ...
    def wait_moves(self) -> None: ...
    def dwell(self, delay: float) -> None: ...
    def flush_step_generation(self) -> None: ...
    def manual_move(self, coord: _Pos | list[float | None], speed: float) -> None: ...
    def get_trapq(self) -> str: ...
    def get_last_move_time(self) -> float: ...



from __future__ import annotations

from typing import TYPE_CHECKING, Protocol

if TYPE_CHECKING:
    from cartographer.interfaces.configuration import Configuration
    from cartographer.interfaces.multiprocessing import TaskExecutor
    from cartographer.interfaces.printer import Mcu, Toolhead
    from cartographer.macros.axis_twist_compensation import AxisTwistCompensationAdapter
    from cartographer.macros.bed_mesh.interfaces import BedMeshAdapter


class Adapters(Protocol):
    config: Configuration
    toolhead: Toolhead
    mcu: Mcu
    axis_twist_compensation: AxisTwistCompensationAdapter | None
    bed_mesh: BedMeshAdapter
    task_executor: TaskExecutor



# https://github.com/Klipper3d/klipper/blob/master/klippy/configfile.py
import configparser
from typing import overload

from klippy import Printer

error = configparser.Error

class ConfigWrapper:
    error: type[configparser.Error]
    printer: Printer
    def __init__(
        self, printer: Printer, fileconfig: configparser.ConfigParser, access_tracking: dict[str, str], section: str
    ) -> None: ...
    def get_printer(self) -> Printer: ...
    def get_name(self) -> str: ...
    def getsection(self, section: str) -> ConfigWrapper: ...
    def has_section(self, section: str) -> bool: ...
    def get_prefix_sections(self, prefix: str) -> list[ConfigWrapper]: ...
    def deprecate(self, option: str, value: str | None = None) -> None: ...
    @overload
    def get(
        self,
        option: str,
        default: str = ...,
        note_valid: bool = True,
    ) -> str: ...
    @overload
    def get(
        self,
        option: str,
        default: None,
        note_valid: bool = True,
    ) -> str | None: ...
    @overload
    def getint(
        self,
        option: str,
        default: int = ...,
        minval: int | None = None,
        maxval: int | None = None,
        note_valid: bool = True,
    ) -> int: ...
    @overload
    def getint(
        self,
        option: str,
        default: None,
        minval: int | None = None,
        maxval: int | None = None,
        note_valid: bool = True,
    ) -> int | None: ...
    @overload
    def getfloat(
        self,
        option: str,
        default: float = ...,
        minval: float | None = None,
        maxval: float | None = None,
        above: float | None = None,
        below: float | None = None,
        note_valid: bool = True,
    ) -> float: ...
    @overload
    def getfloat(
        self,
        option: str,
        default: None,
        minval: float | None = None,
        maxval: float | None = None,
        above: float | None = None,
        below: float | None = None,
        note_valid: bool = True,
    ) -> float | None: ...
    @overload
    def getboolean(
        self,
        option: str,
        default: bool = ...,
        note_valid: bool = True,
    ) -> bool: ...
    @overload
    def getboolean(
        self,
        option: str,
        default: None,
        note_valid: bool = True,
    ) -> bool | None: ...
    def getchoice(
        self,
        option: str,
        choices: dict[str, str],
        default: str = ...,
        note_valid: bool = True,
    ) -> str: ...
    @overload
    def getintlist(
        self,
        option: str,
        default: list[int] = ...,
        sep: str = ",",
        count: int | None = None,
        note_valid: bool = True,
    ) -> list[int]: ...
    @overload
    def getintlist(
        self,
        option: str,
        default: None,
        sep: str = ",",
        count: int | None = None,
        note_valid: bool = True,
    ) -> list[int] | None: ...
    @overload
    def getfloatlist(
        self,
        option: str,
        default: list[float] = ...,
        sep: str = ",",
        count: int | None = None,
        note_valid: bool = True,
    ) -> list[float]: ...
    @overload
    def getfloatlist(
        self,
        option: str,
        default: None,
        sep: str = ",",
        count: int | None = None,
        note_valid: bool = True,
    ) -> list[float] | None: ...

class PrinterConfig:
    def get_printer(self) -> Printer: ...
    def deprecate(self, option: str, value: str | None = None) -> None: ...
    def set(self, section: str, option: str, value: object) -> None: ...
    def remove_section(self, section: str) -> None: ...



from __future__ import annotations

import logging
import math
from dataclasses import dataclass
from typing import TYPE_CHECKING, Protocol

import numpy as np
from typing_extensions import override

from cartographer.interfaces.printer import Endstop, HomingState, Position, ProbeMode, Sample
from cartographer.probe.scan_model import ScanModelSelectorMixin

if TYPE_CHECKING:
    from cartographer.interfaces.configuration import Configuration, ScanModelConfiguration
    from cartographer.interfaces.printer import Mcu, Toolhead
    from cartographer.stream import Session

logger = logging.getLogger(__name__)


class Model(Protocol):
    @property
    def name(self) -> str: ...

    @property
    def z_offset(self) -> float: ...
    def distance_to_frequency(self, distance: float) -> float: ...
    def frequency_to_distance(self, frequency: float) -> float: ...


TRIGGER_DISTANCE = 2.0


@dataclass(frozen=True)
class ScanModeConfiguration:
    x_offset: float
    y_offset: float
    travel_speed: float
    probe_speed: float

    samples: int
    models: dict[str, ScanModelConfiguration]

    @staticmethod
    def from_config(config: Configuration):
        return ScanModeConfiguration(
            x_offset=config.general.x_offset,
            y_offset=config.general.y_offset,
            travel_speed=config.general.travel_speed,
            probe_speed=config.scan.probe_speed,
            samples=config.scan.samples,
            models=config.scan.models,
        )


class ScanMode(ScanModelSelectorMixin, ProbeMode, Endstop):
    """Implementation for Scan mode."""

    @property
    @override
    def offset(self) -> Position:
        return Position(self._config.x_offset, self._config.y_offset, self.probe_height)

    @property
    @override
    def is_ready(self) -> bool:
        return self.has_model()

    @property
    @override
    def last_homing_time(self) -> float:
        return self._last_homing_time

    def __init__(
        self,
        mcu: Mcu,
        toolhead: Toolhead,
        config: ScanModeConfiguration,
    ) -> None:
        super().__init__(config.models)
        self._last_homing_time: float = 0.0
        self._toolhead: Toolhead = toolhead
        self._config: ScanModeConfiguration = config
        self.probe_height: float = TRIGGER_DISTANCE
        self._mcu: Mcu = mcu

        self.last_z_result: float | None = None

    @override
    def get_status(self, eventtime: float) -> object:
        return {
            "current_model": self.get_model().name if self.has_model() else "none",
            "models": ", ".join(self._config.models.keys()),
            "last_z_result": self.last_z_result,
        }

    @override
    def perform_probe(self) -> float:
        if not self._toolhead.is_homed("z"):
            msg = "Z axis must be homed before probing"
            raise RuntimeError(msg)

        dist = self.measure_distance()
        if dist > self.probe_height + 0.5:
            # Safely move downwards
            _ = self._toolhead.z_homing_move(self, speed=self._config.probe_speed)
        elif self._toolhead.get_position().z < self.probe_height:
            self._toolhead.move(z=self.probe_height, speed=self._config.probe_speed)
            self._toolhead.wait_moves()

        delta = self.probe_height - self.measure_distance()

        toolhead_pos = self._toolhead.get_position()
        dist = toolhead_pos.z + delta

        if math.isinf(dist):
            msg = "Toolhead stopped outside model range"
            raise RuntimeError(msg)

        pos = self._toolhead.apply_axis_twist_compensation(Position(toolhead_pos.x, toolhead_pos.y, dist))
        logger.info("probe at %.3f,%.3f is z=%.6f", pos.x, pos.y, pos.z)
        self.last_z_result = pos.z
        return self.last_z_result

    def measure_distance(
        self, *, time: float | None = None, min_sample_count: int | None = None, skip_count: int = 5
    ) -> float:
        model = self.get_model()

        min_sample_count = min_sample_count or self._config.samples
        time = time or self._toolhead.get_last_move_time()

        with self._mcu.start_session(lambda sample: sample.time >= time) as session:
            session.wait_for(lambda samples: len(samples) >= min_sample_count + skip_count)
        samples = session.get_items()[skip_count:]

        dist = float(np.median([model.frequency_to_distance(sample.frequency) for sample in samples]))
        return dist

    def calculate_sample_distance(self, sample: Sample) -> float:
        model = self.get_model()
        return model.frequency_to_distance(sample.frequency)

    @override
    def query_is_triggered(self, print_time: float) -> bool:
        if not self.has_model():
            return True  # No model loaded, assume triggered
        distance = self.measure_distance(time=print_time)
        return distance <= self.get_endstop_position()

    @override
    def get_endstop_position(self) -> float:
        return self.probe_height

    @override
    def home_start(self, print_time: float) -> object:
        trigger_frequency = self.get_model().distance_to_frequency(self.probe_height)
        return self._mcu.start_homing_scan(print_time, trigger_frequency)

    @override
    def on_home_end(self, homing_state: HomingState) -> None:
        if not homing_state.is_homing_z():
            return
        distance = self.measure_distance()
        if not math.isfinite(distance):
            msg = "Toolhead stopped outside model range"
            raise RuntimeError(msg)

        homing_state.set_z_homed_position(distance)
        self._last_homing_time = self._toolhead.get_last_move_time()

    @override
    def home_wait(self, home_end_time: float) -> float:
        return self._mcu.stop_homing(home_end_time)

    def start_session(
        self,
    ) -> Session[Sample]:
        time = self._toolhead.get_last_move_time()
        return self._mcu.start_session(lambda sample: sample.time >= time)



from __future__ import annotations

from typing import Callable

from typing_extensions import override

from cartographer.interfaces.printer import Endstop, HomingAxis, Position, TemperatureStatus, Toolhead


class BacklashCompensatingToolhead(Toolhead):
    def __init__(self, toolhead: Toolhead, backlash: float):
        self.toolhead: Toolhead = toolhead
        self.backlash: float = backlash

    @override
    def move(self, *, x: float | None = None, y: float | None = None, z: float | None = None, speed: float) -> None:
        current_z = self.toolhead.get_position().z
        if z is None or z <= current_z:
            # Moving down or horizontally — no compensation needed
            self.toolhead.move(x=x, y=y, z=z, speed=speed)
            return

        self.toolhead.move(x=x, y=y, z=z + self.backlash, speed=speed)
        self.toolhead.move(x=x, y=y, z=z, speed=speed)

    @override
    def get_position(self) -> Position:
        return self.toolhead.get_position()

    @override
    def get_last_move_time(self) -> float:
        return self.toolhead.get_last_move_time()

    @override
    def wait_moves(self) -> None:
        self.toolhead.wait_moves()

    @override
    def is_homed(self, axis: HomingAxis) -> bool:
        return self.toolhead.is_homed(axis)

    @override
    def get_gcode_z_offset(self) -> float:
        return self.toolhead.get_gcode_z_offset()

    @override
    def z_homing_move(self, endstop: Endstop, *, speed: float) -> float:
        return self.toolhead.z_homing_move(endstop, speed=speed)

    @override
    def set_z_position(self, z: float) -> None:
        self.toolhead.set_z_position(z)

    @override
    def get_axis_limits(self, axis: HomingAxis) -> tuple[float, float]:
        return self.toolhead.get_axis_limits(axis)

    @override
    def manual_probe(self, finalize_callback: Callable[[Position | None], None]) -> None:
        self.toolhead.manual_probe(finalize_callback)

    @override
    def clear_z_homing_state(self) -> None:
        self.toolhead.clear_z_homing_state()

    @override
    def dwell(self, seconds: float) -> None:
        self.toolhead.dwell(seconds)

    @override
    def get_extruder_temperature(self) -> TemperatureStatus:
        return self.toolhead.get_extruder_temperature()

    @override
    def apply_axis_twist_compensation(self, position: Position) -> Position:
        return self.toolhead.apply_axis_twist_compensation(position)



from typing import final

from typing_extensions import override

from cartographer.interfaces.configuration import TouchModelConfiguration
from cartographer.probe.model import ModelSelectorMixin


@final
class TouchModel:
    def __init__(self, config: TouchModelConfiguration) -> None:
        self.config = config

    @property
    def name(self) -> str:
        return self.config.name

    @property
    def z_offset(self) -> float:
        return self.config.z_offset

    @property
    def speed(self) -> float:
        return self.config.speed

    @property
    def threshold(self) -> int:
        return self.config.threshold


class TouchModelSelectorMixin(ModelSelectorMixin[TouchModel, TouchModelConfiguration]):
    @override
    def _create_model(self, config: TouchModelConfiguration) -> TouchModel:
        return TouchModel(config)



from __future__ import annotations

import logging
from abc import ABC, abstractmethod
from typing import Generic, TypeVar

TModel = TypeVar("TModel")
TModelConfig = TypeVar("TModelConfig")

logger = logging.getLogger(__name__)


class ModelSelectorMixin(Generic[TModel, TModelConfig], ABC):
    def __init__(self, models: dict[str, TModelConfig]) -> None:
        self._models: dict[str, TModelConfig] = models
        self._loaded_model: TModel | None = None

    def get_model(self) -> TModel:
        if self._loaded_model is None:
            msg = f"{self._get_type()} model not loaded."
            raise RuntimeError(msg)
        return self._loaded_model

    def has_model(self) -> bool:
        return self._loaded_model is not None

    def load_model(self, name: str):
        model = self._models.get(name, None)
        if model is None:
            msg = f"{self._get_type()} model {name} not found"
            raise RuntimeError(msg)
        self._loaded_model = self._create_model(self._models[name])

    def _get_type(self) -> str:
        """Dynamically retrieve the class name for type."""
        return self.__class__.__name__.removesuffix("Mode")

    @abstractmethod
    def _create_model(self, config: TModelConfig) -> TModel:
        """Override in subclass to build model from config."""
        raise NotImplementedError



from typing import TypedDict

from typing_extensions import NotRequired

class _PrinterObject(TypedDict):
    name: str
    center: NotRequired[list[float]]
    polygon: NotRequired[list[list[float]]]

class _Status(TypedDict):
    objects: list[_PrinterObject]
    excluded_objects: list[_PrinterObject]
    current_object: _PrinterObject | None

class ExcludeObject:
    def get_status(self, eventtime: float | None = ...) -> _Status: ...



# https://github.com/Klipper3d/klipper/blob/master/klippy/extras/thermistor.py

class Thermistor:
    def __init__(self, pullup: float, inline_resistor: float) -> None: ...
    def setup_coefficients_beta(self, t1: float, r1: float, beta: float) -> None: ...
    def calc_temp(self, adc: float) -> float: ...



# pyright: reportUnknownMemberType=false
# pyright: reportUnusedCallResult=false
# pyright: reportMissingParameterType=false
# pyright: reportUnknownParameterType=false
from __future__ import annotations

import sys
from typing import TYPE_CHECKING

import matplotlib.pyplot as plt
from matplotlib.animation import FuncAnimation

from cartographer.macros.bed_mesh.paths.alternating_snake import AlternatingSnakePathGenerator
from cartographer.macros.bed_mesh.paths.random_path import RandomPathGenerator
from cartographer.macros.bed_mesh.paths.snake_path import SnakePathGenerator
from cartographer.macros.bed_mesh.paths.spiral_path import SpiralPathGenerator

if TYPE_CHECKING:
    from matplotlib.text import Annotation

    from cartographer.macros.bed_mesh.interfaces import PathGenerator, Point


def make_grid(nx: int, ny: int, spacing: float) -> list[Point]:
    return [(1 + x * spacing, 1 + y * spacing) for y in range(ny) for x in range(nx)]


def plot_grid_and_path(ax, name: str, grid: list[Point], path: list[Point]):
    # Plot all grid points as light gray background dots
    gx, gy = zip(*grid)
    ax.scatter(gx, gy, color="red", s=20, label="Grid")

    # Plot path as blue line
    px, py = zip(*path)
    ax.plot(px, py, linestyle="-", color="blue", label="Path")

    ax.set_title(name)
    ax.set_aspect("equal")
    ax.grid(True)


def animate_paths(generator: PathGenerator, grid_shapes: list[tuple[int, int]], spacing=5, interval=300):
    fig, axes = plt.subplots(nrows=(len(grid_shapes) + 2) // 3, ncols=3, figsize=(16, 10))
    axes = axes.flatten()

    all_grids: list[list[Point]] = []
    all_paths: list[list[Point]] = []
    arrows: list[Annotation] = []

    # Prepare plots & arrows
    for ax, (nx, ny) in zip(axes, grid_shapes):
        grid = make_grid(nx, ny, spacing)
        path = list(generator.generate_path(grid, (0, nx * spacing + 2), (0, ny * spacing + 2)))
        plot_grid_and_path(ax, f"{nx}×{ny} grid", grid, path)
        all_grids.append(grid)
        all_paths.append(path)

        # Create arrow annotation at start of path
        arrow = ax.annotate(
            "",
            xy=path[0],
            arrowprops=dict(
                arrowstyle="->",
                color="red",
                lw=2,
            ),
        )
        arrows.append(arrow)

    for ax in axes[len(grid_shapes) :]:
        ax.axis("off")

    def update(frame: int):
        for arrow, path in zip(arrows, all_paths):
            i = frame % len(path)  # Loop through frames
            arrow.set_position(xy=(float(path[i - 1][0]), float(path[i - 1][1])))
            arrow.xy = (float(path[i][0]), float(path[i][1]))
        return arrows

    _ = FuncAnimation(fig, update, interval=interval, blit=True)
    plt.tight_layout()
    plt.show()


PATH_STRATEGY_MAP = {
    "snake": SnakePathGenerator,
    "alternating_snake": AlternatingSnakePathGenerator,
    "spiral": SpiralPathGenerator,
    "random": RandomPathGenerator,
}

if __name__ == "__main__":
    path_strategy_type = sys.argv[1]
    path_strategy = PATH_STRATEGY_MAP.get(path_strategy_type)
    if path_strategy is None:
        msg = f"Unknown path strategy: {path_strategy_type}"
        raise ValueError(msg)
    generator = path_strategy("x")

    grid_shapes = [
        (3, 3),
        (4, 4),
        (5, 3),
        (6, 4),
        (10, 11),
        (7, 11),
        (9, 8),
    ]

    animate_paths(generator, grid_shapes)



from __future__ import annotations

import logging
from enum import Enum
from functools import partial
from typing import TYPE_CHECKING, final

from typing_extensions import assert_never, override

from cartographer.interfaces.printer import Macro, MacroParams, Position, Toolhead
from cartographer.macros.utils import get_enum_choice
from cartographer.probe import Probe, ScanModel

if TYPE_CHECKING:
    from cartographer.interfaces.configuration import Configuration

    pass

logger = logging.getLogger(__name__)

DEFAULT_SCAN_MODEL_NAME = "default"


class ScanCalibrateMethod(Enum):
    TOUCH = "touch"
    MANUAL = "manual"


@final
class ScanCalibrateMacro(Macro):
    description = "Run the scan calibration"

    def __init__(self, probe: Probe, toolhead: Toolhead, config: Configuration) -> None:
        self._probe = probe
        self._toolhead = toolhead
        self._config = config

    @override
    def run(self, params: MacroParams) -> None:
        name = params.get("MODEL_NAME", DEFAULT_SCAN_MODEL_NAME)
        method = get_enum_choice(params, "METHOD", ScanCalibrateMethod, default=ScanCalibrateMethod.MANUAL)

        if not self._toolhead.is_homed("x") or not self._toolhead.is_homed("y"):
            msg = "Must home x and y before calibration"
            raise RuntimeError(msg)

        self._toolhead.move(
            x=self._config.bed_mesh.zero_reference_position[0],
            y=self._config.bed_mesh.zero_reference_position[1],
            speed=self._config.general.travel_speed,
        )
        self._toolhead.wait_moves()

        if method == ScanCalibrateMethod.TOUCH:
            return self._run_touch(name)
        elif method == ScanCalibrateMethod.MANUAL:
            return self._run_manual(name)

        assert_never(method)

    def _run_touch(self, name: str) -> None:
        trigger_pos = self._probe.perform_touch()
        pos = self._toolhead.get_position()
        self._toolhead.set_z_position(pos.z - trigger_pos)
        self._calibrate(name)

    def _run_manual(self, name: str) -> None:
        _, z_max = self._toolhead.get_axis_limits("z")
        self._toolhead.set_z_position(z=z_max - 10)

        logger.info("Triggering manual probe, please bring nozzle to 0.1mm above the bed")

        self._toolhead.manual_probe(partial(self._handle_manual_probe, name))

    def _handle_manual_probe(self, name: str, pos: Position | None) -> None:
        if pos is None:
            self._toolhead.clear_z_homing_state()
            return

        # TODO: Should this nozzle offset be customizable?
        # We assume the user will move the nozzle to 0.1mm above the bed
        self._toolhead.set_z_position(0.1)

        self._calibrate(name)

    def _calibrate(self, name: str):
        self._toolhead.move(z=5.5, speed=5)
        self._toolhead.wait_moves()

        with self._probe.scan.start_session() as session:
            session.wait_for(lambda samples: len(samples) > 50)
            self._toolhead.dwell(0.250)
            self._toolhead.move(z=0.1, speed=1)
            self._toolhead.dwell(0.250)
            self._toolhead.wait_moves()
            time = self._toolhead.get_last_move_time()
            session.wait_for(lambda samples: samples[-1].time >= time)
            count = len(session.items)
            session.wait_for(lambda samples: len(samples) >= count + 50)
        self._toolhead.move(z=5, speed=5)

        samples = session.get_items()
        logger.debug("Collected %d samples", len(samples))

        model = ScanModel.fit(name, samples, z_offset=0)
        logger.debug("Scan calibration fitted model: %s", model)

        self._config.save_scan_model(model)
        self._probe.scan.load_model(model.name)
        logger.info(
            """
            Scan model %s has been saved for the current session.
            The SAVE_CONFIG command will update the printer config file and restart the printer.
            """,
            name,
        )



from __future__ import annotations

from cartographer.interfaces.printer import HomingAxis, Position


class MockToolhead:
    """Mock toolhead that tracks movements."""

    def __init__(self):
        self.position: Position = Position(0, 0, 0)
        self.moves: list[tuple[float, float]] = []  # (x, y)
        self.last_move_time: float = 0.0

    def move(self, x: float | None = None, y: float | None = None, z: float | None = None, speed: float | None = None):
        """Track movement commands."""
        del z, speed  # Not used in this mock
        new_x = x if x is not None else self.position.x
        new_y = y if y is not None else self.position.y
        self.last_move_time += 1  # Simulate time passing

        self.position = Position(new_x, new_y, 0)
        self.moves.append((new_x, new_y))

    def wait_moves(self):
        """Mock wait for moves."""
        pass

    def dwell(self, time: float):
        """Mock dwell."""
        self.last_move_time += time

    def get_last_move_time(self) -> float:
        return self.last_move_time

    def get_axis_limits(self, axis: HomingAxis) -> tuple[float, float]:
        """Mock axis limits - return arbitrary limits."""
        del axis
        return (0.0, 100.0)

    def apply_axis_twist_compensation(self, position: Position) -> Position:
        """Mock compensation - just return the same position."""
        return position



from __future__ import annotations

from typing import TYPE_CHECKING, final

import numpy as np
from extras import bed_mesh
from typing_extensions import override

from cartographer.macros.bed_mesh.interfaces import BedMeshAdapter, Polygon

if TYPE_CHECKING:
    from configfile import ConfigWrapper
    from extras.bed_mesh import _Params as BedMeshParams  # pyright: ignore[reportPrivateUsage]

    from cartographer.interfaces.printer import Position

ROUND_DECIMALS = 2


@final
class KlipperBedMesh(BedMeshAdapter):
    def __init__(self, config: ConfigWrapper) -> None:
        self.config = config.getsection("bed_mesh")
        self.bed_mesh = config.get_printer().load_object(self.config, "bed_mesh")
        self.printer = config.get_printer()

    @override
    def get_objects(self) -> list[Polygon]:
        exclude_object = self.printer.lookup_object("exclude_object", None)
        if not exclude_object:
            return []

        objects = exclude_object.get_status().get("objects", [])
        polygons: list[Polygon] = []

        for obj in objects:
            polygon = obj.get("polygon", [])
            points = [(x, y) for x, y in polygon]
            polygons.append(points)

        return polygons

    @override
    def clear_mesh(self) -> None:
        self.bed_mesh.set_mesh(None)

    @override
    def apply_mesh(self, mesh_points: list[Position], profile_name: str | None = None) -> None:
        coords = np.array([p.as_tuple() for p in mesh_points])

        xs_rounded = np.round(coords[:, 0], ROUND_DECIMALS)
        ys_rounded = np.round(coords[:, 1], ROUND_DECIMALS)
        zs = coords[:, 2]

        x_unique = np.unique(xs_rounded)
        y_unique = np.unique(ys_rounded)
        points_per_x = len(x_unique)
        points_per_y = len(y_unique)

        x_indices = {v: i for i, v in enumerate(x_unique)}
        y_indices = {v: i for i, v in enumerate(y_unique)}

        matrix = np.full((points_per_y, points_per_x), np.nan)

        for x, y, z in zip(xs_rounded, ys_rounded, zs):
            xi = x_indices[x]
            yi = y_indices[y]
            matrix[yi, xi] = z

        if np.isnan(matrix).any():
            msg = "Mesh has missing points or inconsistent coordinates"
            raise RuntimeError(msg)

        mesh_params: BedMeshParams = {
            "min_x": float(x_unique[0]),
            "min_y": float(y_unique[0]),
            "max_x": float(x_unique[-1]),
            "max_y": float(y_unique[-1]),
            "x_count": points_per_x,
            "y_count": points_per_y,
            "mesh_x_pps": 0,
            "mesh_y_pps": 0,
            "algo": "bicubic",
            "tension": 0.2,
        }

        mesh = bed_mesh.ZMesh(mesh_params, profile_name)
        try:
            native_matrix = [[float(z) for z in row] for row in matrix.tolist()]
            mesh.build_mesh(native_matrix)
        except bed_mesh.BedMeshError as e:
            raise RuntimeError(str(e)) from e

        self.bed_mesh.set_mesh(mesh)
        if profile_name is not None:
            self.bed_mesh.save_profile(profile_name)



from __future__ import annotations

from enum import Enum
from typing import TYPE_CHECKING, Iterable, TypeVar

if TYPE_CHECKING:
    from cartographer.interfaces.printer import MacroParams

T = TypeVar("T", bound=Enum)


def get_enum_choice(params: MacroParams, option: str, enum_type: type[T], default: T) -> T:
    choice = params.get(option, default=default.value)

    # Convert both the choice and enum values to lowercase for case-insensitive comparison
    lower_choice = str(choice).lower()
    lower_mapping = {str(v.value).lower(): v for v in enum_type}

    if lower_choice not in lower_mapping:
        msg = f"Invalid choice '{choice}' for option '{option}'"
        raise RuntimeError(msg)

    return lower_mapping[lower_choice]


K = TypeVar("K", bound=str)


def get_choice(params: MacroParams, option: str, choices: Iterable[K], default: K) -> K:
    choice = params.get(option, default=default)
    choice_str = choice.lower()

    for k in choices:
        if k.lower() == choice_str:
            return k

    valid_choices = ", ".join(f"'{k.lower()}'" for k in choices)
    msg = f"Invalid choice '{choice}' for option '{option}'. Valid choices are: {valid_choices}"
    raise RuntimeError(msg)


def get_int_tuple(params: MacroParams, option: str, default: tuple[int, int]) -> tuple[int, int]:
    param = params.get(option, default=None)
    if param is None:
        return default
    parts = param.split(",")
    if len(parts) != 2:
        msg = f"Expected two int values for '{option}', got {len(parts)}: {param}"
        raise ValueError(msg)

    return (int(parts[0]), int(parts[1]))


def get_float_tuple(params: MacroParams, option: str, default: tuple[float, float]) -> tuple[float, float]:
    param = params.get(option, default=None)
    if param is None:
        return default
    parts = param.split(",")
    if len(parts) != 2:
        msg = f"Expected two float values for '{option}', got {len(parts)}: {param}"
        raise ValueError(msg)

    return (float(parts[0]), float(parts[1]))



from __future__ import annotations

from typing import TYPE_CHECKING, Iterator, Protocol

from typing_extensions import TypeAlias

if TYPE_CHECKING:
    import numpy as np

    from cartographer.interfaces.printer import Position

Point: TypeAlias = "tuple[float|np.float64, float|np.float64]"
Polygon: TypeAlias = "list[tuple[float, float]]"


class PathGenerator(Protocol):
    def generate_path(
        self, points: list[Point], x_axis_limits: tuple[float, float], y_axis_limits: tuple[float, float]
    ) -> Iterator[Point]: ...


class BedMeshAdapter(Protocol):
    def apply_mesh(self, mesh_points: list[Position], profile_name: str | None = None): ...
    def clear_mesh(self) -> None: ...
    def get_objects(self) -> list[Polygon]: ...



from typing import TYPE_CHECKING, cast

from cartographer.adapters.kalico.adapters import KalicoAdapters
from cartographer.adapters.kalico.integrator import KalicoIntegrator
from cartographer.adapters.klipper.adapters import KlipperAdapters
from cartographer.adapters.klipper.integrator import KlipperIntegrator
from cartographer.runtime.adapters import Adapters
from cartographer.runtime.environment import Environment, detect_environment
from cartographer.runtime.integrator import Integrator

if TYPE_CHECKING:
    from configfile import ConfigWrapper as KlipperConfigWrapper


def init_adapter(config: object) -> Adapters:
    env = detect_environment(config)
    if env == Environment.Klipper:
        from cartographer.adapters.klipper.adapters import KlipperAdapters

        return KlipperAdapters(cast("KlipperConfigWrapper", config))
    if env == Environment.Kalico:
        from cartographer.adapters.kalico.adapters import KalicoAdapters

        return KalicoAdapters(cast("KlipperConfigWrapper", config))

    msg = f"Unsupported environment: {env}"
    raise RuntimeError(msg)


def init_integrator(adapters: Adapters) -> Integrator:
    if isinstance(adapters, KlipperAdapters):
        return KlipperIntegrator(adapters)
    if isinstance(adapters, KalicoAdapters):
        return KalicoIntegrator(adapters)

    msg = "Unsupported adapters"
    raise RuntimeError(msg)



from __future__ import annotations

import logging
from typing import TYPE_CHECKING

import pytest
from typing_extensions import TypeAlias

from cartographer.interfaces.printer import MacroParams, Position, Toolhead
from cartographer.macros.touch import TouchAccuracyMacro, TouchHomeMacro, TouchProbeMacro
from cartographer.probe.touch_mode import TouchMode, TouchModeConfiguration

if TYPE_CHECKING:
    from pytest import LogCaptureFixture
    from pytest_mock import MockerFixture

Probe: TypeAlias = TouchMode


@pytest.fixture
def offset() -> Position:
    return Position(0, 0, 0)


@pytest.fixture
def probe(mocker: MockerFixture, offset: Position) -> Probe:
    mock = mocker.Mock(spec=Probe, autospec=True)
    mock.config = mocker.Mock(spec=TouchModeConfiguration, autospec=True)
    mock.config.move_speed = 42
    mock.offset = offset
    return mock


def test_touch_macro_output(
    mocker: MockerFixture,
    caplog: LogCaptureFixture,
    probe: Probe,
    params: MacroParams,
):
    macro = TouchProbeMacro(probe)
    probe.perform_probe = mocker.Mock(return_value=5.0)

    with caplog.at_level(logging.INFO):
        macro.run(params)

    assert "Result is z=5.000000" in caplog.messages


def test_touch_accuracy_macro_output(
    mocker: MockerFixture,
    caplog: LogCaptureFixture,
    probe: Probe,
    toolhead: Toolhead,
    params: MacroParams,
):
    macro = TouchAccuracyMacro(probe, toolhead)
    params.get_int = mocker.Mock(return_value=10)
    toolhead.get_position = lambda: Position(0, 0, 0)
    params.get_float = mocker.Mock(return_value=1)
    i = -1
    measurements: list[float] = [50 + i * 10 for i in range(10)]

    def mock_probe(**_) -> float:
        nonlocal i
        i += 1
        return measurements[i]

    probe.perform_probe = mock_probe

    with caplog.at_level(logging.INFO):
        macro.run(params)

    assert "touch accuracy results" in caplog.text
    assert "minimum 50" in caplog.text
    assert "maximum 140" in caplog.text
    assert "range 90" in caplog.text
    assert "average 95" in caplog.text
    assert "median 95" in caplog.text
    assert "standard deviation 28" in caplog.text


def test_touch_accuracy_macro_sample_count(
    mocker: MockerFixture,
    caplog: LogCaptureFixture,
    probe: Probe,
    toolhead: Toolhead,
    params: MacroParams,
):
    macro = TouchAccuracyMacro(probe, toolhead)
    params.get_int = mocker.Mock(return_value=3)
    toolhead.get_position = lambda: Position(0, 0, 0)
    params.get_float = mocker.Mock(return_value=1)
    i = -1
    measurements: list[float] = [50 + i * 10 for i in range(10)]

    def mock_probe(**_) -> float:
        nonlocal i
        i += 1
        return measurements[i]

    probe.perform_probe = mock_probe

    with caplog.at_level(logging.INFO):
        macro.run(params)

    assert "touch accuracy results" in caplog.text
    assert "minimum 50" in caplog.text
    assert "maximum 70" in caplog.text
    assert "range 20" in caplog.text
    assert "average 60" in caplog.text
    assert "median 60" in caplog.text
    assert "standard deviation 8" in caplog.text


def test_touch_home_macro_moves(
    mocker: MockerFixture,
    probe: Probe,
    toolhead: Toolhead,
    params: MacroParams,
):
    macro = TouchHomeMacro(probe, toolhead, (10, 10))
    probe.perform_probe = mocker.Mock(return_value=0.1)
    toolhead.get_position = mocker.Mock(return_value=Position(0, 0, 2))
    move_spy = mocker.spy(toolhead, "move")

    macro.run(params)

    assert move_spy.mock_calls == [
        mocker.call(z=4, speed=mocker.ANY),
        mocker.call(x=10, y=10, speed=mocker.ANY),
    ]


def test_touch_home_macro(
    mocker: MockerFixture,
    probe: Probe,
    toolhead: Toolhead,
    params: MacroParams,
):
    # We are at 2, and touch the bed at -0.1.
    height = 2
    trigger = 0.1
    # That means that the bed was further away than we thought,
    # so we need to move the z axis "down".
    expected = height - trigger

    macro = TouchHomeMacro(probe, toolhead, (10, 10))
    probe.perform_probe = mocker.Mock(return_value=trigger)
    toolhead.get_position = mocker.Mock(return_value=Position(0, 0, height))
    set_z_position_spy = mocker.spy(toolhead, "set_z_position")

    macro.run(params)

    assert set_z_position_spy.mock_calls == [mocker.call(expected)]



from __future__ import annotations

from typing import Iterable, Literal, Protocol, TypeVar

from cartographer.interfaces.configuration import (
    BedMeshConfig,
    GeneralConfig,
    ScanConfig,
    ScanModelConfiguration,
    TouchConfig,
    TouchModelConfiguration,
)

K = TypeVar("K", bound=str)


def get_choice(params: ParseConfigWrapper, option: str, choices: Iterable[K], default: K) -> K:
    choice = params.get_str(option, default=default)
    choice_str = choice.lower()

    for k in choices:
        if k.lower() == choice_str:
            return k

    valid_choices = ", ".join(f"'{k.lower()}'" for k in choices)
    msg = f"Invalid choice '{choice}' for option '{option}'. Valid choices are: {valid_choices}"
    raise RuntimeError(msg)


class ParseConfigWrapper(Protocol):
    def get_name(self) -> str: ...
    def get_str(self, option: str, default: str) -> str: ...
    def get_optional_str(self, option: str) -> str | None: ...
    def get_float(
        self, option: str, default: float, minimum: float | None = None, maximum: float | None = None
    ) -> float: ...
    def get_required_float(self, option: str) -> float: ...
    def get_required_float_list(self, option: str, count: int | None = None) -> list[float]: ...
    def get_int(self, option: str, default: int) -> int: ...
    def get_required_int_list(self, option: str, count: int | None = None) -> list[int]: ...
    def get_bool(self, option: str, default: bool) -> bool: ...


T = TypeVar("T")


def list_to_tuple(lst: list[T]) -> tuple[T, T]:
    if len(lst) != 2:
        msg = f"Expected a list of length 2, got {len(lst)}"
        raise ValueError(msg)
    return (lst[0], lst[1])


def parse_general_config(wrapper: ParseConfigWrapper) -> GeneralConfig:
    return GeneralConfig(
        x_offset=wrapper.get_required_float("x_offset"),
        y_offset=wrapper.get_required_float("y_offset"),
        z_backlash=wrapper.get_float("z_backlash", default=0.05, minimum=0),
        travel_speed=wrapper.get_float("travel_speed", default=50, minimum=1),
        macro_prefix=wrapper.get_optional_str("macro_prefix"),
        verbose=wrapper.get_bool("verbose", default=False),
    )


_directions: list[Literal["x", "y"]] = ["x", "y"]
_paths: list[Literal["snake", "alternating_snake", "spiral", "random"]] = [
    "snake",
    "alternating_snake",
    "spiral",
    "random",
]


def parse_scan_config(wrapper: ParseConfigWrapper, models: dict[str, ScanModelConfiguration]) -> ScanConfig:
    return ScanConfig(
        samples=20,
        models=models,
        probe_speed=wrapper.get_float("probe_speed", default=5, minimum=0.1),
        mesh_runs=wrapper.get_int("mesh_runs", default=1),
        mesh_direction=get_choice(wrapper, "mesh_direction", _directions, default="x"),
        mesh_height=wrapper.get_float("mesh_height", default=4, minimum=1),
        mesh_path=get_choice(wrapper, "mesh_path", _paths, default="snake"),
    )


def parse_touch_config(wrapper: ParseConfigWrapper, models: dict[str, TouchModelConfiguration]) -> TouchConfig:
    samples = wrapper.get_int("samples", default=5)
    return TouchConfig(
        samples=samples,
        max_samples=wrapper.get_int("max_samples", default=samples * 2),
        models=models,
    )


def parse_bed_mesh_config(wrapper: ParseConfigWrapper) -> BedMeshConfig:
    return BedMeshConfig(
        mesh_min=list_to_tuple(wrapper.get_required_float_list("mesh_min", count=2)),
        mesh_max=list_to_tuple(wrapper.get_required_float_list("mesh_max", count=2)),
        probe_count=list_to_tuple(wrapper.get_required_int_list("probe_count", count=2)),
        speed=wrapper.get_float("speed", default=50, minimum=1),
        horizontal_move_z=wrapper.get_float("horizontal_move_z", default=4, minimum=1),
        adaptive_margin=wrapper.get_float("adaptive_margin", default=5, minimum=0),
        zero_reference_position=list_to_tuple(wrapper.get_required_float_list("zero_reference_position", count=2)),
    )


def parse_scan_model_config(wrapper: ParseConfigWrapper) -> ScanModelConfiguration:
    return ScanModelConfiguration(
        name=wrapper.get_name(),
        coefficients=wrapper.get_required_float_list("coefficients"),
        domain=list_to_tuple(wrapper.get_required_float_list("domain", count=2)),
        z_offset=wrapper.get_float("z_offset", default=0),
    )


def parse_touch_model_config(wrapper: ParseConfigWrapper) -> TouchModelConfiguration:
    return TouchModelConfiguration(
        name=wrapper.get_name(),
        threshold=wrapper.get_int("threshold", default=100),
        speed=wrapper.get_float("speed", default=50, minimum=1),
        z_offset=wrapper.get_float("z_offset", default=0, maximum=0),
    )



from __future__ import annotations

import random
from typing import TYPE_CHECKING, Iterator, Literal, final

import numpy as np
from typing_extensions import override

from cartographer.macros.bed_mesh.interfaces import PathGenerator

if TYPE_CHECKING:
    from cartographer.macros.bed_mesh.interfaces import Point

MIN_DIST = 10.0


@final
class RandomPathGenerator(PathGenerator):
    def __init__(self, main_direction: Literal["x", "y"]):
        del main_direction

    @override
    def generate_path(
        self,
        points: list[Point],
        x_axis_limits: tuple[float, float],
        y_axis_limits: tuple[float, float],
    ) -> Iterator[Point]:
        del x_axis_limits, y_axis_limits

        if not points:
            return iter([])

        remaining = points[:]
        current = random.choice(remaining)
        yield current
        remaining.remove(current)

        while remaining:
            current_arr = np.array(current)
            dists = np.array([np.linalg.norm(np.array(p) - current_arr) for p in remaining])

            if np.all(dists == 0):
                next_point = random.choice(remaining)
            else:
                weights = dists / dists.sum()
                next_point = random.choices(remaining, weights=weights, k=1)[0]

            yield next_point
            remaining.remove(next_point)
            current = next_point



from __future__ import annotations

import math
from collections import defaultdict
from typing import TYPE_CHECKING, Any, Iterator, Literal, cast

import numpy as np
from typing_extensions import TypeAlias

if TYPE_CHECKING:
    from numpy.typing import NDArray

    from cartographer.macros.bed_mesh.interfaces import Point

Vec: TypeAlias = "NDArray[np.float64]"


def cluster_points(points: list[Point], axis: Literal["x", "y"], tol: float = 1e-3) -> list[list[Point]]:
    # axis to cluster on:
    # if main_direction = "x", cluster on y (index 1)
    # if main_direction = "y", cluster on x (index 0)
    cluster_index = 1 if axis == "x" else 0
    sort_index = 0 if axis == "x" else 1

    clusters: dict[float, list[Point]] = defaultdict(list)
    for p in points:
        key = round(p[cluster_index] / tol)
        clusters[key].append(p)

    sorted_keys = sorted(clusters.keys())

    rows: list[list[Point]] = []
    for key in sorted_keys:
        row_points = clusters[key]
        row_points.sort(key=lambda pt: pt[sort_index])
        rows.append(row_points)

    return rows


def arc_points(
    center: Vec, radius: float, start_angle_deg: float, span_deg: float, max_dev: float = 0.1
) -> Iterator[Point]:
    cx, cy = center
    if radius == 0:
        yield (cx, cy)
        return

    max_dev = min(max_dev, radius)  # Avoid domain error in arccos
    start_rad = np.deg2rad(start_angle_deg)
    span_rad = np.deg2rad(span_deg)

    d_theta = np.arccos(1 - max_dev / radius)
    n_points = max(1, int(np.ceil(abs(span_rad) / d_theta)))
    thetas = cast("np.ndarray[Any, np.dtype[np.float64]]", start_rad + np.linspace(0, span_rad, n_points + 1))  # pyright: ignore[reportExplicitAny]

    xs = cx + radius * np.cos(thetas)
    ys = cy + radius * np.sin(thetas)

    yield from zip(xs, ys)


def perpendicular(v: Vec, ccw: bool = True) -> Vec:
    return np.array([-v[1], v[0]]) if ccw else np.array([v[1], -v[0]])


def angle_deg(v: Vec) -> float:
    return math.degrees(math.atan2(v[1], v[0]))


def normalize(v: Vec) -> Vec:
    norm = np.linalg.norm(v)
    return v / norm if norm != 0 else v


def row_direction(row: list[Point]) -> Vec:
    if len(row) < 2:
        msg = "Need at least two points to determine direction"
        raise ValueError(msg)
    p0: Vec = np.array(row[0], dtype=float)
    p1: Vec = np.array(row[1], dtype=float)
    dir_vec = p1 - p0
    return dir_vec / np.linalg.norm(dir_vec)  # normalized
